from typing import Union

import abjad

from .extract_trivial_tuplets import (
    extract_trivial_tuplets as extract_trivial_tuplets_function,
)


def prettify_rewrite_meter(
    selection: abjad.Selection,
    meter: Union[abjad.Meter, abjad.TimeSignature],
    *,
    fuse_across_groups_of_beats: bool = True,
    fuse_quadruple_meter: bool = True,
    fuse_triple_meter: bool = True,
    extract_trivial_tuplets: bool = True,
    split_quadruple_meter: bool = True,
) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function fuses pitched leaves according to the rules shown below,
    improving the default output of |abjad.Meter.rewrite_meter()|.

    Basic usage:
        Meters whose denominators are a crotchet or longer get tied notes
        within a beat after |abjad.Meter.rewrite_meter()| when they are at
        an offset ``denominator / 4``, so a rhythm such as  ``denominator / 4``
        ``denominator / 2`` ``denominator / 4`` becomes ``denominator / 4``
        ``denominator / 4`` ``~`` ``denominator / 4`` ``denominator / 4``. This
        function looks for those specific cases and fuses them, generating an
        output which is often more readable.

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'16 d'8 e'16 f'16 g'16 a'8 b'8 c''16 d''16"
        ... )
        >>> meter = abjad.Meter((3, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'16
                d'16
                ~
                d'16
                e'16
                f'16
                g'16
                a'8
                b'8
                c''16
                d''16
            }

        ..  figure:: ../_images/prettify_rewrite_meter-vlnd7l5fb7s.png

        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'16
                d'8
                e'16
                f'16
                g'16
                a'8
                b'8
                c''16
                d''16
            }

        ..  figure:: ../_images/prettify_rewrite_meter-e7vfnese0ut.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.mutate.prettify_rewrite_meter(staff[:], meter)

    Other examples:
        The rhythm of the leaves just before and after the two leaves to be
        fused can be different than ``denominator / 4``, as the function
        searches for logical ties of specific length and offset, and its
        surroundings do not matter.

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'32 d'32 e'8 f'16 "
        ...     r"\times 2/3 {g'32 a'32 b'32} c''8 "
        ...     r"r16 r32. d''64 e''8 f''32 g''32"
        ... )
        >>> meter = abjad.Meter((3, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'32
                d'32
                e'16
                ~
                e'16
                f'16
                \times 2/3
                {
                    g'32
                    a'32
                    b'32
                }
                c''16
                ~
                c''16
                r16
                r32.
                d''64
                e''16
                ~
                e''16
                f''32
                g''32
            }

        ..  figure:: ../_images/prettify_rewrite_meter-kw09gse2zxj.png

        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'32
                d'32
                e'8
                f'16
                \times 2/3
                {
                    g'32
                    a'32
                    b'32
                }
                c''8
                r16
                r32.
                d''64
                e''8
                f''32
                g''32
            }

        ..  figure:: ../_images/prettify_rewrite_meter-u5gmtdippsa.png

    ``fuse_across_groups_of_beats``:
        By default, this function also fuses rhythms of type
        ``denominator / 2`` ``denominator / 2`` ``~`` ``denominator / 2``
        ``denominator / 2``, becoming ``denominator / 2`` ``denominator``
        ``denominator / 2``. This is only applied when the meter's structure
        has a depth of 2, which is the case for meters with numerators equal to
        or larger than ``5``.

        >>> staff = abjad.Staff(r"\time 6/4 c'8 d'4 e'4 f'4 g'4 a'4 b'8")
        >>> meter = abjad.Meter((6, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 6/4
                c'8
                d'8
                ~
                d'8
                e'8
                ~
                e'8
                f'8
                ~
                f'8
                g'8
                ~
                g'8
                a'8
                ~
                a'8
                b'8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-tqi4p0u8qog.png

        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 6/4
                c'8
                d'4
                e'4
                f'8
                ~
                f'8
                g'4
                a'4
                b'8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-riif1glyqpo.png

        to disable this behaviour, set the optional keyword argument
        ``fuse_across_groups_of_beats`` to ``False``.

        >>> staff = abjad.Staff(r"\time 6/4 c'8 d'4 e'4 f'4 g'4 a'4 b'8")
        >>> meter = abjad.Meter((6, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> auxjad.mutate.prettify_rewrite_meter(
        ...     staff[:],
        ...     meter,
        ...     fuse_across_groups_of_beats=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 6/4
                c'8
                d'8
                ~
                d'8
                e'8
                ~
                e'8
                f'8
                ~
                f'8
                g'8
                ~
                g'8
                a'8
                ~
                a'8
                b'8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-ki5xbiteij.png

    |abjad.Meter| with ``increase_monotonic=True``:
        The fused notes will respect the beat structures of such meters, even
        when ``increase_monotonic`` is set to the non-default value ``True``.
        Compare the outputs below.

        >>> staff = abjad.Staff(r"\time 7/4 c'8 d'4 e'4 f'4 g'4 a'4 b'4 c''8")
        >>> meter = abjad.Meter((7, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 7/4
                c'8
                d'4
                e'4
                f'8
                ~
                f'8
                g'4
                a'8
                ~
                a'8
                b'4
                c''8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-bud0jhkvvl.png

        >>> staff = abjad.Staff(r"\time 7/4 c'8 d'4 e'4 f'4 g'4 a'4 b'4 c''8")
        >>> meter = abjad.Meter((7, 4), increase_monotonic=True)
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 7/4
                c'8
                d'4
                e'8
                ~
                e'8
                f'4
                g'8
                ~
                g'8
                a'4
                b'4
                c''8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-47y86pbwwv5.png

    Multiple measures at once:
        This function can take handle multiple measures at once, as long as
        they share the same meter.

        >>> staff = abjad.Staff(
        ...     r"\time 5/8 c'16 d'8 e'8 f'8 g'8 a'16 ~ "
        ...     r"a'16 b'8 c''8 d''8 e''8 f''16"
        ... )
        >>> meter = abjad.Meter((5, 8))
        >>> for measure in abjad.select(staff[:]).group_by_measure():
        ...     abjad.mutate.rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/8
                c'16
                d'16
                ~
                d'16
                e'16
                ~
                e'16
                f'16
                ~
                f'16
                g'16
                ~
                g'16
                a'16
                ~
                a'16
                b'16
                ~
                b'16
                c''16
                ~
                c''16
                d''16
                ~
                d''16
                e''16
                ~
                e''16
                f''16
            }

        ..  figure:: ../_images/prettify_rewrite_meter-8jdzmvf9yl.png

        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/8
                c'16
                d'8
                e'8
                f'16
                ~
                f'16
                g'8
                a'16
                ~
                a'16
                b'8
                c''8
                d''16
                ~
                d''16
                e''8
                f''16
            }

        ..  figure:: ../_images/prettify_rewrite_meter-pcn8x9hr6bb.png

    Multiple measures:
        Similarly to |abjad.Meter.rewrite_meter()|, this function accepts
        selections of multiple measures:

        >>> staff = abjad.Staff(
        ...     r"\time 4/4 c'8 d'4 e'4 f'4 g'8 | "
        ...     r"a'8 b'4 c''8 d''16 e''4 f''8."
        ... )
        >>> meter = abjad.Meter((4, 4))
        >>> for measure in abjad.select(staff[:]).group_by_measure():
        ...     abjad.mutate.rewrite_meter(measure, meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'8
                ~
                d'8
                e'8
                ~
                e'8
                f'8
                ~
                f'8
                g'8
                a'8
                b'8
                ~
                b'8
                c''8
                d''16
                e''8.
                ~
                e''16
                f''8.
            }

        ..  figure:: ../_images/prettify_rewrite_meter-s8fg7a2k0tr.png

        >>> for measure in abjad.select(staff[:]).group_by_measure():
        ...     auxjad.mutate.prettify_rewrite_meter(measure, meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'4
                e'8
                ~
                e'8
                f'4
                g'8
                a'8
                b'4
                c''8
                d''16
                e''8.
                ~
                e''16
                f''8.
            }

        ..  figure:: ../_images/prettify_rewrite_meter-rgd7ok7fkq.png

    Multiple measures with different meters:
        If the measures have different meters, they can be passed on
        individually using :func:`zip()` as shown below.

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'8 d'4 e'4 f'16 g'16 | "
        ...     r"\time 4/4 a'8 b'4 c''8 d''16 e''4 f''8."
        ... )
        >>> meters = [abjad.Meter((3, 4)), abjad.Meter((4, 4))]
        >>> for meter, measure in zip(
        ...     meters,
        ...     abjad.select(staff[:]).group_by_measure(),
        ... ):
        ...     abjad.mutate.rewrite_meter(measure, meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8
                d'8
                ~
                d'8
                e'8
                ~
                e'8
                f'16
                g'16
                \time 4/4
                a'8
                b'8
                ~
                b'8
                c''8
                d''16
                e''8.
                ~
                e''16
                f''8.
            }

        ..  figure:: ../_images/prettify_rewrite_meter-o2izz0m7s9k.png

        >>> for meter, measure in zip(
        ...     meters,
        ...     abjad.select(staff[:]).group_by_measure(),
        ... ):
        ...     auxjad.mutate.prettify_rewrite_meter(measure, meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8
                d'4
                e'4
                f'16
                g'16
                \time 4/4
                a'8
                b'4
                c''8
                d''16
                e''8.
                ~
                e''16
                f''8.
            }

        ..  figure:: ../_images/prettify_rewrite_meter-zh89kk66zon.png

    ``fuse_quadruple_meter``:
        This function also takes care of two special cases, namely quadruple
        and triple meters. By default, it will fuse leaves in quadruple meters
        across beats 1 and 2, and across beats 3 and 4 (as long as they fulfil
        the other requirements of duration and offset).

        >>> staff = abjad.Staff(r"\time 4/4 c'8 d'4 e'4 f'4 g'8")
        >>> meter = abjad.Meter((4, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'4
                e'8
                ~
                e'8
                f'4
                g'8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-nap4bbf7mxe.png

        Set ``fuse_quadruple_meter`` to ``False`` to disable this behaviour.

        >>> staff = abjad.Staff(r"\time 4/4 c'8 d'4 e'4 f'4 g'8")
        >>> meter = abjad.Meter((4, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> auxjad.mutate.prettify_rewrite_meter(
        ...     staff[:],
        ...     meter,
        ...     fuse_quadruple_meter=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'8
                ~
                d'8
                e'8
                ~
                e'8
                f'8
                ~
                f'8
                g'8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-juipg9nzna.png

    ``fuse_triple_meter``:
        In the case of triple meters, it will fuse leaves across any beat as
        long as the previously mentioned conditions of offset and duration are
        met.

        >>> staff = abjad.Staff(r"\time 3/4 c'8 d'4 e'4 f'8")
        >>> meter = abjad.Meter((3, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> auxjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8
                d'4
                e'4
                f'8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-4wg3grpb94p.png

        Similarly to the example before, set ``fuse_triple_meter`` to ``False``
        to disable this behaviour.

        >>> staff = abjad.Staff(r"\time 3/4 c'8 d'4 e'4 f'8")
        >>> meter = abjad.Meter((3, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> auxjad.mutate.prettify_rewrite_meter(
        ...     staff[:],
        ...     meter,
        ...     fuse_triple_meter=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8
                d'8
                ~
                d'8
                e'8
                ~
                e'8
                f'8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-l16ostzscta.png

    ``extract_trivial_tuplets``:
        By default, this function extracts the contents of tuples that consist
        solely of rests, or solely of tied notes and chords.

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'4 ~ c'8} \times 2/3 {d'8 r4} "
        ...     r"\times 2/3 {r8 r8 r8} \times 2/3 {<e' g'>8 ~ <e' g'>4}"
        ... )
        >>> meter = abjad.Meter((4, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> abjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \times 2/3
                {
                    d'8
                    r4
                }
                r4
                <e' g'>4
            }

        ..  figure:: ../_images/prettify_rewrite_meter-a72jx4fc1xd.png

        Set ``extract_trivial_tuplets`` to ``False`` to disable this behaviour.

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'4 ~ c'8} \times 2/3 {d'8 r4} "
        ...     r"\times 2/3 {r8 r8 r8} \times 2/3 {<e' g'>8 ~ <e' g'>4}"
        ... )
        >>> meter = abjad.Meter((4, 4))
        >>> abjad.mutate.rewrite_meter(staff[:], meter)
        >>> abjad.mutate.prettify_rewrite_meter(
        ...     staff[:]
        ...     meter,
        ...     extract_trivial_tuplets=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'4.
                }
                \times 2/3
                {
                    d'8
                    r4
                }
                \times 2/3
                {
                    r4.
                }
                \times 2/3
                {
                    <e' g'>4.
                }
            }

        ..  figure:: ../_images/prettify_rewrite_meter-v9q0ka94qcd.png

    ``split_quadruple_meter``
        When applying |abjad.Meter.rewrite_meter()| to a selection with
        quadruple meter and without using a deeper ``boundary_depth`` than the
        default, the resulting notation will often have leaves crossing the
        third beat of the measure, as shown below.

        >>> staff = abjad.Staff(
        ...     r"c'4 d'2 r4"
        ...     r"e'4. f'2 g'8"
        ...     r"a'4. b'4. c''4"
        ...     r"d''16 e''8. f''4. g''4 a''8"
        ... )
        >>> meter = abjad.Meter((4, 4))
        >>> for measure in abjad.select(staff[:]).group_by_measure():
        ...     abjad.mutate.rewrite_meter(measure, meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'2
                r4
                e'4.
                f'8
                ~
                f'4.
                g'8
                a'4.
                b'4.
                c''4
                d''16
                e''8.
                f''4.
                g''8
                ~
                g''8
                a''8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-1wvhrjife1i.png

        This function tests those leaves against a series of rules, splitting
        them when the tests fails. In the case shown above, the first two bars
        are very easy to read rhythmically, but the third and fourth are less
        so. This is due to the dotted crotchet, which starts off a beat,
        crossing the third beat of the measure. This function will split these
        sort of leaves as shown below.

        >>> abjad.mutate.prettify_rewrite_meter(staff[:], meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'2
                r4
                e'4.
                f'8
                ~
                f'4.
                g'8
                a'4.
                b'8
                ~
                b'4
                c''4
                d''16
                e''8.
                f''4
                ~
                f''8
                g''4
                a''8
            }
        ..  figure:: ../_images/prettify_rewrite_meter-56dy04wjzg.png

        Set ``split_quadruple_meter`` to ``False`` to disable this behaviour.

        >>> staff = abjad.Staff(
        ...     r"c'4 d'2 r4"
        ...     r"e'4. f'2 g'8"
        ...     r"a'4. b'4. c''4"
        ...     r"d''16 e''8. f''4. g''4 a''8"
        ... )
        >>> meter = abjad.Meter((4, 4))
        >>> for measure in abjad.select(staff[:]).group_by_measure():
        ...     abjad.mutate.rewrite_meter(measure, meter)
        >>> abjad.mutate.prettify_rewrite_meter(
        ...     staff[:]
        ...     meter,
        ...     split_quadruple_meter=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'2
                r4
                e'4.
                f'8
                ~
                f'4.
                g'8
                a'4.
                b'4.
                c''4
                d''16
                e''8.
                f''4.
                g''4
                a''8
            }

        ..  figure:: ../_images/prettify_rewrite_meter-ww1x0zsxlnd.png

    ..  tip::

        Use :func:`auxjad.auto_rewrite_meter()` to automatically apply
        |abjad.Meter.rewrite_meter()| and
        |auxjad.mutate.prettify_rewrite_meter()| to a container with multiple
        time signatures.

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("first argument must be 'abjad.Selection'")
    if not selection.leaves().are_contiguous_logical_voice():
        raise ValueError("first argument must be contiguous logical voice")
    if not isinstance(meter, (abjad.Meter, abjad.TimeSignature)):
        raise TypeError("argument must be 'abjad.Meter' or 'abjad.TimeSignature'")
    if not isinstance(fuse_across_groups_of_beats, bool):
        raise TypeError("'fuse_across_groups_of_beats' must be 'bool'")
    if not isinstance(fuse_quadruple_meter, bool):
        raise TypeError("'fuse_quadruple_meter' must be 'bool'")
    if not isinstance(fuse_triple_meter, bool):
        raise TypeError("'fuse_triple_meter' must be 'bool'")
    if not isinstance(extract_trivial_tuplets, bool):
        raise TypeError("'extract_trivial_tuplets' must be 'bool'")
    if not isinstance(split_quadruple_meter, bool):
        raise TypeError("'split_quadruple_meter' must be 'bool'")

    if isinstance(meter, abjad.TimeSignature):
        meter = abjad.Meter(meter.pair)

    logical_ties = selection.logical_ties(pitched=True)
    if len(logical_ties) == 0:
        return

    first_leaf = selection.leaf(0)
    initial_offset = abjad.get.timespan(first_leaf).start_offset
    base = 1 / meter.denominator

    def _merge_indicators_then_fuse(logical_tie):
        last_indicators = abjad.get.indicators(logical_tie[-1])
        initial_indicators = abjad.get.indicators(logical_tie[0])
        initial_indicators_types = tuple(type(indicator) for indicator in initial_indicators)
        abjad.mutate.fuse(logical_tie)
        for indicator in last_indicators:
            if isinstance(
                indicator,
                (
                    abjad.Dynamic,
                    abjad.StopSlur,
                    abjad.StopGroup,
                    abjad.StopHairpin,
                    abjad.StopTextSpan,
                    abjad.StopTrillSpan,
                    abjad.StopPianoPedal,
                    abjad.StopPhrasingSlur,
                ),
            ):
                if not isinstance(indicator, initial_indicators_types):
                    abjad.attach(indicator, logical_tie[0])

    for logical_tie in logical_ties.filter_duration("==", base / 2):
        offset = abjad.get.timespan(logical_tie).start_offset
        offset -= initial_offset
        offset %= base
        if offset == base / 4:
            _merge_indicators_then_fuse(logical_tie)

    if fuse_across_groups_of_beats:
        for logical_tie in logical_ties.filter_duration("==", base):
            offset = abjad.get.timespan(logical_tie).start_offset
            offset -= initial_offset
            offset %= meter.duration
            offset_mod = offset % base
            if offset_mod == base / 2:
                if not offset + base / 2 in meter.depthwise_offset_inventory[1]:
                    _merge_indicators_then_fuse(logical_tie)

    if fuse_quadruple_meter and meter.numerator == 4:
        for logical_tie in logical_ties.filter_duration("==", base):
            offset = abjad.get.timespan(logical_tie).start_offset
            offset -= initial_offset
            offset %= meter.duration
            offset_mod = offset % base
            if offset_mod == base / 2:
                if not offset + base / 2 in (
                    abjad.Offset(0, 1),
                    abjad.Offset(2 * base),
                    abjad.Offset(4 * base),
                ):
                    _merge_indicators_then_fuse(logical_tie)

    if fuse_triple_meter and meter.numerator == 3:
        for logical_tie in logical_ties.filter_duration("==", base):
            offset = abjad.get.timespan(logical_tie).start_offset
            offset -= initial_offset
            offset %= meter.duration
            offset_mod = offset % base
            if offset_mod == base / 2:
                if not offset + base / 2 in (
                    abjad.Offset(0, 1),
                    abjad.Offset(3 * base),
                ):
                    _merge_indicators_then_fuse(logical_tie)

    logical_ties = selection.logical_ties()  # splitting not only pitched
    if split_quadruple_meter and meter.numerator == 4:
        half_point_offset = abjad.Offset(2 * base)
        for logical_tie in logical_ties:
            offset0 = abjad.get.timespan(logical_tie).start_offset
            offset0 -= initial_offset
            offset0 %= meter.duration
            if offset0 == abjad.Offset(0, 1):
                # do not split things like c'1, c'2.. r8, c'2. r4, etc.
                continue
            offset1 = abjad.get.timespan(logical_tie).stop_offset
            offset1 -= initial_offset
            offset1 %= meter.duration
            if offset1 == abjad.Offset(0, 1):
                offset1 = abjad.Offset(meter.duration)
            if offset0 % abjad.Offset(base) == abjad.Offset(0, 1) and offset1 % abjad.Offset(
                base
            ) == abjad.Offset(0, 1):
                # do not split things like r4 c'2.
                continue
            if offset0 < half_point_offset < offset1:
                if offset1 == abjad.Offset(meter.duration) and len(logical_tie) == 1:
                    # do not split r8 c'2.., r16 c'2..., etc.
                    break
                if any(
                    abjad.get.duration(leaf) == abjad.Duration(2 * base) for leaf in logical_tie
                ):
                    # do not split r8 c'8~c'2 r4, r16 c'8.~c'2~c'8 r8, etc.
                    break
                # do not split tuplets
                result = selection.tuplets()
                result = abjad.select(abjad.select(_).logical_ties() for _ in result)
                result = result.flatten()
                if logical_tie in result:
                    break
                duration = abjad.Duration(half_point_offset - offset0)
                abjad.mutate.split(logical_tie, [duration])

    if extract_trivial_tuplets:
        extract_trivial_tuplets_function(selection)



from typing import Optional, Union

import abjad

from .. import get
from .extract_trivial_tuplets import (
    extract_trivial_tuplets as extract_trivial_tuplets_function,
)
from .merge_partial_tuplets import (
    merge_partial_tuplets as merge_partial_tuplets_function,
)
from .prettify_rewrite_meter import (
    prettify_rewrite_meter as prettify_rewrite_meter_function,
)


def auto_rewrite_meter(
    container: abjad.Container,
    meter_list: Optional[
        list[
            Union[
                abjad.Meter,
                abjad.TimeSignature,
            ]
        ]
    ] = None,
    *,
    prettify_rewrite_meter: bool = True,
    extract_trivial_tuplets: bool = True,
    fuse_across_groups_of_beats: bool = True,
    fuse_quadruple_meter: bool = True,
    fuse_triple_meter: bool = True,
    boundary_depth: Optional[int] = None,
    maximum_dot_count: Optional[int] = None,
    rewrite_tuplets: bool = True,
    merge_partial_tuplets: bool = True,
    split_quadruple_meter: bool = True,
) -> None:
    r"""Mutates an input container (of type |abjad.Container| or child class)
    in place and has no return value; this function takes every measure of a
    container, detects its time signature, and apply both
    |abjad.Meter.rewrite_meter()| and
    |auxjad.mutate.prettify_rewrite_meter()| to it.

    Basic usage:
        For this example, the following container will be mutated:

        >>> staff = abjad.Staff(
        ...     r"c'16 d'8 e'16 f'8 g'4 a'4 b'8 "
        ...     r"c'16 d'4. e'16 f'8 g'4 a'16 b'16"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                d'8
                e'16
                f'8
                g'4
                a'4
                b'8
                c'16
                d'4.
                e'16
                f'8
                g'4
                a'16
                b'16
            }

        ..  figure:: ../_images/auto_rewrite_meter-xyx2wh7ufer.png

        Abjad's |abjad.Meter.rewrite_meter()| mutates an |abjad.Selection|
        of a measure, improving its notation.

        >>> for measure in abjad.select(staff[:]).group_by_measure():
        ...     abjad.Meter.rewrite_meter(measure, abjad.Meter((4, 4)))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                d'16
                ~
                d'16
                e'16
                f'8
                g'8
                ~
                g'8
                a'8
                ~
                a'8
                b'8
                c'16
                d'8.
                ~
                d'8.
                e'16
                f'8
                g'8
                ~
                g'8
                a'16
                b'16
            }

        ..  figure:: ../_images/auto_rewrite_meter-7fn2uj2xupb.png

        This function mutates an |abjad.Container| (or child class),
        identifying the implied meters of each measure and applying both
        |abjad.Meter.rewrite_meter()| and
        |auxjad.mutate.prettify_rewrite_meter()| to it. See the documentation
        of the latter for a detailed explanation of what it does.

        Applying |auxjad.mutate.auto_rewrite_meter()| to the same initial
        container shown in the first figure above outputs:

        >>> staff = abjad.Staff(
        ...     r"c'16 d'8 e'16 f'8 g'4 a'4 b'8 "
        ...     r"c'16 d'4. e'16 f'8 g'4 a'16 b'16"
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                d'8
                e'16
                f'8
                g'8
                ~
                g'8
                a'4
                b'8
                c'16
                d'8.
                ~
                d'8.
                e'16
                f'8
                g'4
                a'16
                b'16
            }

        ..  figure:: ../_images/auto_rewrite_meter-ahdaggaiqbc.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.auto_rewrite_meter(staff)
        >>> abjad.mutate.auto_rewrite_meter(staff)

    Time signature changes:
        It automatically handles time signature changes.

        >>> staff = abjad.Staff(
        ...     r"c'16 d'8 e'16 f'8 g'4 a'4 b'8 "
        ...     r"\time 6/8 b'4 c''4 r4 "
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                d'8
                e'16
                f'8
                g'8
                ~
                g'8
                a'4
                b'8
                \time 6/8
                b'4
                c''8
                ~
                c''8
                r4
            }

        ..  figure:: ../_images/auto_rewrite_meter-08sckfp19vil.png

    ``prettify_rewrite_meter``:
        By default, this function invokes both |abjad.Meter.rewrite_meter()|
        and |auxjad.mutate.prettify_rewrite_meter()|.

        >>> staff = abjad.Staff(
        ...     r"c'16 d'8 e'16 f'8 g'4 a'4 b'8 "
        ...     r"c'16 d'8 e'16 f'8 g'4 a'4 b'8"
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                d'8
                e'16
                f'8
                g'8
                ~
                g'8
                a'4
                b'8
                c'16
                d'8
                e'16
                f'8
                g'8
                ~
                g'8
                a'4
                b'8
            }

        ..  figure:: ../_images/auto_rewrite_meter-vbytyszlkng.png

        Set ``prettify_rewrite_meter`` to ``False`` to not invoke
        |auxjad.mutate.prettify_rewrite_meter()|.

        >>> staff = abjad.Staff(
        ...     r"c'16 d'8 e'16 f'8 g'4 a'4 b'8 "
        ...     r"c'16 d'4. e'16 f'8 g'4 a'16 b'16"
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(
        ...     staff,
        ...     prettify_rewrite_meter=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                d'16
                ~
                d'16
                e'16
                f'8
                g'8
                ~
                g'8
                a'8
                ~
                a'8
                b'8
                c'16
                d'8.
                ~
                d'8.
                e'16
                f'8
                g'8
                ~
                g'8
                a'16
                b'16
            }

        ..  figure:: ../_images/auto_rewrite_meter-64wse58hvko.png

    ``meter_list``:
        When no ``meter_list`` is supplied, this function detects the time
        signature of each measure and uses those when rewritting it:

        >>> staff = abjad.Staff(
        ...     r"\time 7/4 c'8 d'4 e'4 f'4 g'4 a'4 b'4 c''8 "
        ...     r"\time 5/4 d''8 e''4 f''4 g''4 a''4 b''8"
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 7/4
                c'8
                d'4
                e'4
                f'8
                ~
                f'8
                g'4
                a'8
                ~
                a'8
                b'4
                c''8
                \time 5/4
                d''8
                e''4
                f''4
                g''8
                ~
                g''8
                a''4
                b''8
            }

        ..  figure:: ../_images/auto_rewrite_meter-l4xnpevp3z.png

        To use a custom list of meters (one for each measure), set
        ``meter_list`` to a :obj:`list` of |abjad.Meter|'s or
        |abjad.TimeSignature|'s.

        >>> staff = abjad.Staff(
        ...     r"\time 7/4 c'8 d'4 e'4 f'4 g'4 a'4 b'4 c''8 "
        ...     r"\time 5/4 d''8 e''4 f''4 g''4 a''4 b''8"
        ... )
        >>> meter_list = [abjad.Meter((7, 4), increase_monotonic=True),
        ...               abjad.Meter((5, 4), increase_monotonic=True),
        ...               ]
        >>> auxjad.mutate.auto_rewrite_meter(staff, meter_list=meter_list)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 7/4
                c'8
                d'4
                e'8
                ~
                e'8
                f'4
                g'8
                ~
                g'8
                a'4
                b'4
                c''8
                \time 5/4
                d''8
                e''4
                f''8
                ~
                f''8
                g''4
                a''4
                b''8
            }

        ..  figure:: ../_images/auto_rewrite_meter-uqif4i8tqxk.png

    Number of measures:
        This function handles a container with any number of measures and any
        number of time signature changes:

        >>> staff = abjad.Staff(
        ...     r"\time 3/4 c'8 d'4 e'4 f'8 "
        ...     r"\time 5/8 g'4 a'4 r8 "
        ...     r"\time 6/8 b'4 c''4 r4 "
        ...     r"\time 4/4 d''8 e''4 f''8 g''16 a''4 r8."
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8
                d'4
                e'4
                f'8
                \time 5/8
                g'4
                a'8
                ~
                a'8
                r8
                \time 6/8
                b'4
                c''8
                ~
                c''8
                r4
                \time 4/4
                d''8
                e''4
                f''8
                g''16
                a''8.
                ~
                a''16
                r8.
            }

        ..  figure:: ../_images/auto_rewrite_meter-hkhtqnqita.png

    ``extract_trivial_tuplets``:
        By default, tuplets filled with rests or tied notes or chords are
        extracted:

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'4 ~ c'8} \times 2/3 {d'8 r4} "
        ...     r"\times 2/3 {r8 r8 r8} \times 2/3 {<e' g'>8 ~ <e' g'>4}"
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \times 2/3
                {
                    d'8
                    r4
                }
                r4
                <e' g'>4
            }

        ..  figure:: ../_images/auto_rewrite_meter-nq6t6qwka7a.png

        Set ``extract_trivial_tuplets`` to ``False`` to disable this behaviour.

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'4 ~ c'8} \times 2/3 {d'8 r4} "
        ...     r"\times 2/3 {r8 r8 r8} \times 2/3 {<e' g'>8 ~ <e' g'>4}"
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(
        ...     staff,
        ...     extract_trivial_tuplets=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'4.
                }
                \times 2/3
                {
                    d'8
                    r4
                }
                \times 2/3
                {
                    r4.
                }
                \times 2/3
                {
                    <e' g'>4.
                }
            }

        ..  figure:: ../_images/auto_rewrite_meter-ssnsui7o9cc.png

    ``merge_partial_tuplets``:
        By default, consecutive partial tuplets with the same ratio that sum up
        to an assignable duration will be merged together:

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'2 d'1}"
        ...     r"\times 2/3 {e'2} \times 2/3 {f'1}"
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'2
                    d'1
                }
                \times 2/3
                {
                    e'2
                    f'1
                }
            }

        ..  figure:: ../_images/auto_rewrite_meter-ty72t5wvc1.png

        Set ``merge_partial_tuplets`` to ``False`` to disable this behaviour.

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'2 d'1}"
        ...     r"\times 2/3 {e'2} \times 2/3 {f'1}"
        ... )
        >>> auxjad.mutate.auto_rewrite_meter(
        ...     staff,
        ...     merge_partial_tuplets=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'2
                    d'1
                }
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    e'2
                }
                \tweak edge-height #'(0.7 . 0)
                \times 2/3
                {
                    f'1
                }
            }

        ..  figure:: ../_images/auto_rewrite_meter-4rouf819bjb.png

    ..  note::

        This function also accepts the arguments ``boundary_depth``,
        ``maximum_dot_count``, and ``rewrite_tuplets``, which are passed on to
        |abjad.Meter.rewrite_meter()|, and ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, ``fuse_triple_meter``, and
        ``split_quadruple_meter``, which are passed on to
        |auxjad.mutate.prettify_rewrite_meter()|. ``merge_partial_tuplets``
        is used to invoke |auxjad.mutate.merge_partial_tuplets()| See the
        documentation of these functions for more details on these arguments.

    ..  warning::

        Setting ``boundary_depth`` to a value equal to or larger than ``1``
        will automatically disable ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, and ``fuse_triple_meter``, regardless of
        their values. This is because when any of those arguments is ``True``,
        |auxjad.mutate.prettify_rewrite_meter()| will fuse across beats,
        which goes against the purpose of using ``boundary_depth``. Compare the
        results below. In the first case, simply applying
        |auxjad.mutate.prettify_rewrite_meter()| with no arguments results in
        some logical ties being tied across beats.

        >>> staff = abjad.Staff(r"\time 4/4 c'4. d'4. e'4 f'8 g'4 a'4 b'4.")
        >>> meter = abjad.Meter((4, 4))
        >>> for measure in abjad.select(staff[:]).group_by_measure():
        ...     abjad.mutate.rewrite_meter(measure, meter, boundary_depth=1)
        >>> for measure in abjad.select(staff[:]).group_by_measure():
        ...     auxjad.mutate.prettify_rewrite_meter(measure, meter)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                ~
                c'8
                d'8
                ~
                d'4
                e'4
                f'8
                g'4
                a'8
                ~
                a'8
                b'8
                ~
                b'4
            }

        ..  figure:: ../_images/auto_rewrite_meter-cf09ysj16fo.png

        By automatically setting all ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, and  ``fuse_triple_meter` to ``False`` when
        ``boundary_depth`` is equal to or larger than ``1``, this function will
        not fuse those leaves against the required boundary depth.

        >>> staff = abjad.Staff(r"\time 4/4 c'4. d'4. e'4 f'8 g'4 a'4 b'4.")
        >>> auxjad.mutate.auto_rewrite_meter(staff, boundary_depth=1)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                ~
                c'8
                d'8
                ~
                d'4
                e'4
                f'8
                g'8
                ~
                g'8
                a'8
                ~
                a'8
                b'8
                ~
                b'4
            }

        ..  figure:: ../_images/auto_rewrite_meter-mm9xvmaqwfj.png
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("first positional argument must be 'abjad.Container' or child class")
    if meter_list is not None:
        if not isinstance(meter_list, list):
            raise TypeError(
                "'meter_list' must be a 'list' of 'abjad.Meter' or 'abjad.TimeSignature'"
            )
        else:
            for meter in meter_list:
                if not isinstance(meter, (abjad.Meter, abjad.TimeSignature)):
                    raise TypeError(
                        "elements of 'meter_list' must be 'abjad.Meter' or 'abjad.TimeSignature'"
                    )
                if isinstance(meter, abjad.TimeSignature):
                    meter = abjad.Meter(meter.pair)
    if not isinstance(prettify_rewrite_meter, bool):
        raise TypeError("'prettify_rewrite_meter' must be 'bool'")
    if not isinstance(fuse_across_groups_of_beats, bool):
        raise TypeError("'fuse_across_groups_of_beats' must be 'bool'")
    if not isinstance(fuse_quadruple_meter, bool):
        raise TypeError("'fuse_quadruple_meter' must be 'bool'")
    if not isinstance(fuse_triple_meter, bool):
        raise TypeError("'fuse_triple_meter' must be 'bool'")
    if boundary_depth is not None:
        if not isinstance(boundary_depth, int):
            raise TypeError("'boundary_depth' must be 'int'")
    if maximum_dot_count is not None:
        if not isinstance(maximum_dot_count, int):
            raise TypeError("'maximum_dot_count' must be 'int'")
    if not isinstance(rewrite_tuplets, bool):
        raise TypeError("'rewrite_tuplets' must be 'bool'")
    if not isinstance(merge_partial_tuplets, bool):
        raise TypeError("'merge_partial_tuplets' must be 'bool'")
    if not isinstance(split_quadruple_meter, bool):
        raise TypeError("'split_quadruple_meter' must be 'bool'")

    if merge_partial_tuplets:
        merge_partial_tuplets_function(container[:])
    if extract_trivial_tuplets:
        extract_trivial_tuplets_function(container[:])

    if meter_list is None:
        time_signatures = get.time_signature_list(
            container,
            do_not_use_none=True,
        )
        meter_list = [abjad.Meter(ts.pair) for ts in time_signatures]
    measures = abjad.select(container[:]).group_by_measure()
    for meter, measure in zip(meter_list, measures):
        if isinstance(measure[0], abjad.MultimeasureRest):
            continue
        abjad.Meter.rewrite_meter(
            measure,
            meter,
            boundary_depth=boundary_depth,
            maximum_dot_count=maximum_dot_count,
            rewrite_tuplets=rewrite_tuplets,
        )
    if prettify_rewrite_meter:
        measures = abjad.select(container[:]).group_by_measure()
        for meter, measure in zip(meter_list, measures):
            if boundary_depth is None or boundary_depth < 1:
                prettify_rewrite_meter_function(
                    measure,
                    meter,
                    fuse_across_groups_of_beats=fuse_across_groups_of_beats,
                    fuse_quadruple_meter=fuse_quadruple_meter,
                    fuse_triple_meter=fuse_triple_meter,
                    extract_trivial_tuplets=False,
                    split_quadruple_meter=split_quadruple_meter,
                )
            else:
                prettify_rewrite_meter_function(
                    measure,
                    meter,
                    fuse_across_groups_of_beats=False,
                    fuse_quadruple_meter=False,
                    fuse_triple_meter=False,
                    extract_trivial_tuplets=False,
                    split_quadruple_meter=False,
                )



from collections.abc import Iterable
from typing import Union

import abjad

from .. import get
from .close_container import close_container
from .rests_to_multimeasure_rest import rests_to_multimeasure_rest


def sync_containers(
    containers: Union[
        Iterable[abjad.Container],
        abjad.Score,
    ],
    *,
    use_multimeasure_rests: bool = True,
    adjust_last_time_signature: bool = True,
) -> None:
    r"""Mutates two or more input containers in place and has no return value;
    this function finds the longest container among the inputs and adds rests
    to all the shorter ones, making them the same length. Input argument can
    be a single |abjad.Score| with multiple containers, or an iterable with
    elements of type |abjad.Container| or child classes.

    Basic usage:
        Input two or more containers. This function will fill the shortest ones
        with rests ensuring all their lengths become the same.

        >>> staff1 = abjad.Staff(r"\time 4/4 g'2.")
        >>> staff2 = abjad.Staff(r"\time 4/4 c'1")
        >>> abjad.show(staff1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                g'2.
            }

        ..  figure:: ../_images/sync_containers-H7jNwGEtFQ.png

        >>> abjad.show(staff2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
            }

        ..  figure:: ../_images/sync_containers-EjKqQCinPy.png

        >>> auxjad.mutate.sync_containers([staff1, staff2])
        >>> abjad.show(staff1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                g'2.
                r4
            }

        ..  figure:: ../_images/sync_containers-akcdf8t9e5.png

        >>> abjad.show(staff2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
            }

        ..  figure:: ../_images/sync_containers-l7tru1tjoli.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.sync_containers([container1, container2])
        >>> abjad.mutate.sync_containers([container1, container2])

    ..  note::

        When using |abjad.Container|'s, all time signatures in the output will
        be commented out with ``%%%.`` This is because Abjad only applies time
        signatures to containers that belong to a |abjad.Staff|. The present
        function works with either |abjad.Container| and |abjad.Staff|.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> abjad.show(container)

        ..  docs::

            {
                %%% \time 3/4 %%%
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/sync_containers-9sl3dnd2uwn.png

        >>> staff = abjad.Staff([container])
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/sync_containers-08v2pv2tmqqn.png

    Containers of same size:
        If all containers have the same size, no modification is applied.

        >>> container1 = abjad.Staff(r"\time 3/4 g'2.")
        >>> container2 = abjad.Staff(r"\time 3/4 c'2.")
        >>> auxjad.mutate.sync_containers([container1, container2])
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 3/4
                g'2.
            }

        ..  figure:: ../_images/sync_containers-e0yszxejbh.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
            }

        ..  figure:: ../_images/sync_containers-2cgt4zds3h7.png

    Underfull containers:
        By default, this function closes the longest container by rewriting the
        time signature of its last measure if necessary (if it is underfull),
        and uses multi-measure rests whenever possible.

        >>> container1 = abjad.Staff(r"\time 4/4 g'1 | f'4")
        >>> container2 = abjad.Staff(r"\time 4/4 c'1")
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                g'1
                f'4
            }

        ..  figure:: ../_images/sync_containers-udnAjAZtkw.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
            }

        ..  figure:: ../_images/sync_containers-cy8dL4UUWV.png

        >>> auxjad.mutate.sync_containers([container1, container2])
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                g'1
                \time 1/4
                f'4
            }

        ..  figure:: ../_images/sync_containers-nztndgecrof.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                \time 1/4
                R1 * 1/4
            }

        ..  figure:: ../_images/sync_containers-iaag195ty1d.png

    ``adjust_last_time_signature``:
        To allow containers to be left open (with underfull measures), set the
        keyword argument ``adjust_last_time_signature`` to ``False``.

        >>> container1 = abjad.Staff(r"\time 4/4 g'1 | f'4")
        >>> container2 = abjad.Staff(r"\time 4/4 c'1")
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                g'1
                f'4
            }

        ..  figure:: ../_images/sync_containers-uldd5Vni8a.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
            }

        ..  figure:: ../_images/sync_containers-eTCrjHBr39.png

        >>> auxjad.mutate.sync_containers(
        ...     [container1, container2],
        ...     adjust_last_time_signature=False,
        ... )
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                g'1
                f'4
            }

        ..  figure:: ../_images/sync_containers-37iesjp4dqs.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                r4
            }

        ..  figure:: ../_images/sync_containers-lqm4itxlwu.png

    ``use_multimeasure_rests``:
        To disable multi-measure rests, set the keyword argument
        ``use_multimeasure_rests`` to ``False``.

        >>> container1 = abjad.Staff(r"\time 4/4 g'1 | f'4")
        >>> container2 = abjad.Staff(r"\time 4/4 c'1")
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                g'1
                \time 1/4
                f'4
            }

        ..  figure:: ../_images/sync_containers-Mbmjf7JLH8.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
            }

        ..  figure:: ../_images/sync_containers-3RZwROEIt1.png

        >>> auxjad.mutate.sync_containers(
        ...     [container1, container2],
        ...     use_multimeasure_rests=False,
        ... )
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                g'1
                \time 1/4
                f'4
            }

        ..  figure:: ../_images/sync_containers-rhagiugx42o.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                \time 1/4
                r4
            }

        ..  figure:: ../_images/sync_containers-oss03t1qnf8.png

    Adjusting last time signatures:
        When adjusting the last time signature, this function will maintain the
        same time effective signature for as long as possible and only add a
        new one at the last measure if its duration is shorter.

        >>> container1 = abjad.Staff(r"\time 7/4 a'1 ~ a'2.")
        >>> container2 = abjad.Staff(r"\time 3/4 c'2.")
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 7/4
                a'1
                ~
                a'2.
            }

        ..  figure:: ../_images/sync_containers-g61gTyfZGV.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
            }

        ..  figure:: ../_images/sync_containers-UZvmef5N16.png

        >>> auxjad.mutate.sync_containers([container1, container2])
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 7/4
                a'1
                ~
                a'2.
            }

        ..  figure:: ../_images/sync_containers-dQnq2ASkTu.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                R1 * 3/4
                \time 1/4
                R1 * 1/4
            }

        ..  figure:: ../_images/sync_containers-jhx0r9skgwi.png

    Multiple input containers:
        This function can take an arbitrary number of containers.

        >>> container1 = abjad.Staff(r"\time 4/4 c'1 | g'4")
        >>> container2 = abjad.Staff(r"\time 4/4 c'1 | g'2")
        >>> container3 = abjad.Staff(r"\time 4/4 c'1 | g'2.")
        >>> container4 = abjad.Staff(r"\time 4/4 c'1")
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                g'4
            }

        ..  figure:: ../_images/sync_containers-epvaWYPYT7.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                g'2
            }

        ..  figure:: ../_images/sync_containers-Yja5mx93Fi.png

        >>> abjad.show(container3)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                g'2.
            }

        ..  figure:: ../_images/sync_containers-GTtAEELheA.png

        >>> abjad.show(container4)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
            }

        ..  figure:: ../_images/sync_containers-VCFpuJX7ID.png

        >>> containers = [container1,
        ...               container2,
        ...               container3,
        ...               container4,
        ...               ]
        >>> auxjad.mutate.sync_containers(containers)
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                \time 3/4
                g'4
                r2
            }

        ..  figure:: ../_images/sync_containers-1wbsyvks33r.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                \time 3/4
                g'2
                r4
            }

        ..  figure:: ../_images/sync_containers-td1whqky24b.png

        >>> abjad.show(container3)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                \time 3/4
                g'2.
            }

        ..  figure:: ../_images/sync_containers-g07scyil9jh.png

        >>> abjad.show(container4)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'1
                \time 3/4
                R1 * 3/4
            }

        ..  figure:: ../_images/sync_containers-8b6vn3azaom.png

    Single input |abjad.Score|:
        This function can also take a single |abjad.Score| instead of
        multiple |abjad.Container|'s or |abjad.Staff|'s.

        >>> staff1 = abjad.Staff(r"\time 3/8 c'4. | d'4")
        >>> staff2 = abjad.Staff(r"\time 3/8 c'4. | d'8")
        >>> staff3 = abjad.Staff(r"\time 3/8 c'4. | d'16")
        >>> staff4 = abjad.Staff(r"\time 3/8 c'4.")
        >>> score = abjad.Score([staff1,
        ...                      staff2,
        ...                      staff3,
        ...                      staff4,
        ...                      ])
        >>> auxjad.mutate.sync_containers(score)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \time 3/8
                    c'4.
                    \time 1/4
                    d'4
                }
                \new Staff
                {
                    \time 3/8
                    c'4.
                    \time 1/4
                    d'8
                    r8
                }
                \new Staff
                {
                    \time 3/8
                    c'4.
                    \time 1/4
                    d'16
                    r8.
                }
                \new Staff
                {
                    \time 3/8
                    c'4.
                    \time 1/4
                    R1 * 1/4
                }
            >>

        ..  figure:: ../_images/sync_containers-0g0651fs0luq.png

    Time signature changes:
        The containers can be of different length, can have different time
        signatures, and can contain time signature changes as well.

        >>> container1 = abjad.Staff(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Staff(r"\time 3/4 a2. \time 4/4 c'4")
        >>> container3 = abjad.Staff(r"\time 5/4 g''1 ~ g''4")
        >>> container4 = abjad.Staff(r"\time 6/8 c'2")
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/sync_containers-qAt4FSyAIl.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 3/4
                a2.
                \time 4/4
                c'4
            }

        ..  figure:: ../_images/sync_containers-EKrqsCDVNs.png

        >>> abjad.show(container3)

        ..  docs::

            \new Staff
            {
                \time 5/4
                g''1
                ~
                g''4
            }

        ..  figure:: ../_images/sync_containers-lw5fknvt4o.png

        >>> abjad.show(container4)

        ..  docs::

            \new Staff
            {
                \time 6/8
                c'2
            }

        ..  figure:: ../_images/sync_containers-iXGajGKBZl.png

        >>> containers = [container1,
        ...               container2,
        ...               container3,
        ...               container4,
        ...               ]
        >>> auxjad.mutate.sync_containers(containers)
        >>> abjad.show(container1)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
                \time 1/4
                R1 * 1/4
            }

        ..  figure:: ../_images/sync_containers-mec52wgbrz9.png

        >>> abjad.show(container2)

        ..  docs::

            \new Staff
            {
                \time 3/4
                a2.
                \time 2/4
                c'4
                r4
            }

        ..  figure:: ../_images/sync_containers-33odhzqyo6r.png

        >>> abjad.show(container3)

        ..  docs::

            \new Staff
            {
                \time 5/4
                g''1
                ~
                g''4
            }

        ..  figure:: ../_images/sync_containers-s7rmadmd1f.png

        >>> abjad.show(container4)

        ..  docs::

            \new Staff
            {
                \time 6/8
                c'2
                r4
                \time 2/4
                R1 * 1/2
            }

        ..  figure:: ../_images/sync_containers-msu922pcn6e.png

    Polymetric notation:
        In order to enable polymetric notation in LilyPond, the command
        ``"\enablePolymeter"`` must be added to the layout block. Below is a
        full example of how this can be accomplished using Abjad.

        >>> container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"\time 3/4 a2. \time 4/4 c'4")
        >>> container3 = abjad.Container(r"\time 5/4 g''1 ~ g''4")
        >>> container4 = abjad.Container(r"\time 6/8 c'2")
        >>> containers = [container1,
        ...               container2,
        ...               container3,
        ...               container4,
        ...               ]
        >>> auxjad.mutate.sync_containers(containers)
        >>> staves = [abjad.Staff([container1]),
        ...           abjad.Staff([container2]),
        ...           abjad.Staff([container3]),
        ...           abjad.Staff([container4]),
        ...           ]
        >>> score = abjad.Score(staves)
        >>> lilypond_file = abjad.LilyPondFile.new()
        >>> score_block = abjad.Block(name='score')
        >>> layout_block = abjad.Block(name='layout')
        >>> score_block.items.append(score)
        >>> score_block.items.append(layout_block)
        >>> lilypond_file.items.append(score_block)
        >>> layout_block.items.append(r'\enablePolymeter')
        >>> abjad.show(lilypond_file)

        ..  docs::

            \score { %! abjad.LilyPondFile._get_formatted_blocks()
                \new Score
                <<
                    \new Staff
                    {
                        {
                            \time 4/4
                            c'4
                            d'4
                            e'4
                            f'4
                            \time 1/4
                            R1 * 1/4
                        }
                    }
                    \new Staff
                    {
                        {
                            \time 3/4
                            a2.
                            \time 2/4
                            c'4
                            r4
                        }
                    }
                    \new Staff
                    {
                        {
                            \time 5/4
                            g''1
                            ~
                            g''4
                        }
                    }
                    \new Staff
                    {
                        {
                            \time 6/8
                            c'2
                            r4
                            \time 2/4
                            R1 * 1/2
                        }
                    }
                >>
                \layout {
                    \enablePolymeter
                }
            } %! abjad.LilyPondFile._get_formatted_blocks()

        ..  figure:: ../_images/sync_containers-1lbrepesgil.png

    ..  error::

        If one or more containers is malformed, i.e. it has an underfilled
        measure before a time signature change, the function raises a
        :exc:`ValueError` exception.

        >>> container1 = abjad.Container(r"\time 4/4 g'1 | f'4")
        >>> container2 = abjad.Container(r"\time 5/4 c'1 | \time 4/4 d'4")
        >>> auxjad.mutate.sync_containers([container1, container2])
        ValueError: at least one 'container' is malformed, with an underfull
        measure preceding a time signature change
    """
    if not isinstance(containers, (Iterable, abjad.Score)):
        raise TypeError("argument must be 'abjad.Score' or iterable of 'abjad.Container's")
    if isinstance(containers, abjad.Score):
        containers = containers[:]
    for container in containers:
        if not isinstance(container, abjad.Container):
            raise TypeError("argument must be 'abjad.Score' or iterable of 'abjad.Container's")
        if not abjad.select(container).leaves().are_contiguous_logical_voice():
            raise ValueError("argument must each be contiguous logical voice")
        try:
            get.selection_is_full(container[:])
        except ValueError as err:
            raise ValueError(
                "at least one 'container' is malformed, with an "
                "underfull measure preceding a time signature "
                "change"
            ) from err
    if not isinstance(use_multimeasure_rests, bool):
        raise TypeError("'use_multimeasure_rests' must be 'bool'")
    if not isinstance(adjust_last_time_signature, bool):
        raise TypeError("'adjust_last_time_signature' must be 'bool'")

    durations = [abjad.get.duration(container[:]) for container in containers]
    max_duration = max(durations)
    for container, duration in zip(containers, durations):
        duration_difference = max_duration - duration
        if duration_difference > abjad.Duration(0):
            # handling duration left in the last measure, if any
            if not get.selection_is_full(container[:]):
                duration_left = get.underfull_duration(container[:])
                underfull_rests_duration = min(
                    duration_difference,
                    duration_left,
                )
                underfull_rests = abjad.LeafMaker()(
                    None,
                    underfull_rests_duration,
                )
                duration_difference -= underfull_rests_duration
                container.extend(underfull_rests)
                if duration_difference == abjad.Duration(0) and adjust_last_time_signature:
                    close_container(container)
                if duration_difference == abjad.Duration(0):
                    continue
            # finding out last effective time signature
            for leaf in abjad.select(container).leaves()[::-1]:
                effective_time_signature = abjad.get.effective(
                    leaf,
                    abjad.TimeSignature,
                )
                if effective_time_signature is not None:
                    break
            else:
                effective_time_signature = abjad.TimeSignature((4, 4))
            # creating new measures for any leftover duration
            measure_duration = effective_time_signature.duration
            while duration_difference > measure_duration:
                rests = abjad.LeafMaker()(None, measure_duration)
                duration_difference -= measure_duration
                container.extend(rests)
            if duration_difference > abjad.Duration(0):
                rests = abjad.LeafMaker()(None, duration_difference)
                if adjust_last_time_signature:
                    rests_time_signature = abjad.TimeSignature(
                        duration_difference,
                    )
                    rests_time_signature.simplify_ratio()
                    if rests_time_signature != effective_time_signature:
                        abjad.attach(rests_time_signature, rests[0])
                container.extend(rests)
            if use_multimeasure_rests:
                rests_to_multimeasure_rest(container[:])
        else:
            # closing longest container if necessary
            if adjust_last_time_signature and not get.selection_is_full(container[:]):
                close_container(container)



import random
from typing import Optional, Union

import abjad

from .. import get, mutate


class Phaser:
    r"""Takes an |abjad.Container| (or child class) as input and outputs an
    |abjad.Selection| with leaves shifted by a fixed amount. Subsequent calls
    apply further shifts.

    Basic usage:
        Calling the object will return an |abjad.Selection| generated by
        the phasing process. Each call of the object will shift the contents of
        the input container by a fixed amount given by :attr:`step_size`. Note
        that, by default, the first call outputs the initial container, with
        subsequent calls phasing it.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Phaser-VjQH8XsyZ6.png

        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        )
        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Phaser-953j4yo4hlo.png

        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
                ~
                f'8.
                c'16
            }

        ..  figure:: ../_images/Phaser-3l0a96vbfbi.png

        The property :attr:`current_window` can be used to access the current
        window without moving the head forwards.

        >>> notes = phaser.current_window()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
                ~
                f'8.
                c'16
            }

        ..  figure:: ../_images/Phaser-pn2qf2z6vzr.png

    :attr:`process_on_first_call`:
        The very first call will output the input container without processing
        it. To disable this behaviour and phase on the very first call,
        initialise the class with the keyword argument
        :attr:`process_on_first_call` set to ``True``.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        process_on_first_call=True,
        ...                        )
        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
                ~
                f'8.
                c'16
            }

        ..  figure:: ../_images/Phaser-fx0ywlw5u25.png

    :attr:`step_size`:
        The argument :attr:`step_size` is used to step sizes for the phasing
        process. It takes a :obj:`tuple` or an |abjad.Duration|.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 8),
        ...                        )
        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Phaser-zfxas2f5o87.png

        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'4
                e'8
                ~
                e'8
                f'4
                c'8
            }

        ..  figure:: ../_images/Phaser-2e9dzvji4wg.png

    Using as iterator:
        The instances of :class:`Phaser` can also be used as an iterator, which
        can then be used in a for loop to phase through a full cycle. Note that
        unlike the methods :meth:`output_n` and :meth:`output_all`, time
        signatures are added to each window returned by the shuffler. Use the
        function |auxjad.mutate.remove_repeated_time_signatures()| to clean
        the output when using :class:`Phaser` in this way.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 ~ e'2.")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 4),
        ...                        )
        >>> staff = abjad.Staff()
        >>> for window in phaser:
        ...     staff.append(window)
        >>> auxjad.mutate.remove_repeated_time_signatures(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                ~
                e'2.
                d'4
                e'2
                ~
                e'2
                c'4
                e'2.
                ~
                e'4
                c'4
                d'4
                e'2.
                c'4
                d'4
                e'4
                e'2
                c'4
                d'4
                e'2
                e'4
                c'4
                d'4
                e'2.
            }

        ..  figure:: ../_images/Phaser-jd1h0w8pbpp.png

    Arguments and properties:
        This class can take many optional keyword arguments during its
        creation. :attr:`max_steps` sets the maximum number of steps that can
        be applied for a single call, ranging between ``1`` and the input value
        (default is also ``1``); if :attr:`max_steps` is set to a value larger
        than 1, the number of steps at every call will be randomly chosen.
        :attr:`forward_bias` sets the chance of the process moving forward
        instead of backwards. It should range from ``0.0`` to ``1.0`` (default
        ``1.0``, which means the process can only move forwards. A value of
        ``0.5`` gives 50% chance of moving forwards while a value of ``0.0``
        will result in the process moving only backwards). By default, when a
        logical tie is split in between windows, any unterminated ties will be
        removed; set :attr:`remove_unterminated_ties` to ``False`` to disable
        this behaviour. The properties :attr:`boundary_depth`,
        :attr:`maximum_dot_count`, and :attr:`rewrite_tuplets` are passed as
        arguments to |abjad.Meter.rewrite_meter()|, see its documentation for
        more information. By default, calling the object will first return the
        original container and subsequent calls will process it; set
        :attr:`process_on_first_call` to ``True`` and the looping process will
        be applied on the very first call. Setting the property
        :attr:`omit_time_signatures` to ``True`` will remove all time
        signatures from the output (``False`` by default).

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(5, 8),
        ...                        max_steps=2,
        ...                        forward_bias=0.2,
        ...                        remove_unterminated_ties=True,
        ...                        omit_time_signatures=True,
        ...                        boundary_depth=0,
        ...                        maximum_dot_count=1,
        ...                        rewrite_tuplets=False,
        ...                        process_on_first_call=True,
        ...                        )
        >>> phaser.step_size
        5/8
        >>> phaser.max_steps
        2
        >>> phaser.forward_bias
        0.2
        >>> phaser.remove_unterminated_ties
        True
        >>> phaser.omit_time_signatures
        True
        >>> phaser.boundary_depth
        0
        >>> phaser.maximum_dot_count
        1
        >>> phaser.rewrite_tuplets
        False
        >>> phaser.process_on_first_call
        True

        Use the properties below to change these values after initialisation.

        >>> phaser.step_size = (1, 4)
        >>> phaser.max_steps = 3
        >>> phaser.forward_bias = 0.8
        >>> phaser.remove_unterminated_ties = False
        >>> phaser.omit_time_signatures = False
        >>> phaser.boundary_depth = 1
        >>> phaser.maximum_dot_count = 2
        >>> phaser.rewrite_tuplets = True
        >>> phaser.process_on_first_call = False
        >>> phaser.step_size
        1/4
        >>> phaser.max_steps
        3
        >>> phaser.forward_bias
        0.8
        >>> phaser.remove_unterminated_ties
        False
        >>> phaser.omit_time_signatures
        False
        >>> phaser.boundary_depth
        1
        >>> phaser.maximum_dot_count
        2
        >>> phaser.rewrite_tuplets
        True
        >>> phaser.process_on_first_call
        False

    Setting :attr:`forward_bias` to ``0.0``:
        Set :attr:`forward_bias` to ``0.0`` to move backwards instead of
        forwards (default is ``1.0``).

        >>> container = abjad.Container(r"\time 3/8 c'8 d'8 e'8")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        )
        >>> notes = phaser.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'8
                d'8
                e'8
                c'16
                d'8
                e'8
                c'16
                d'8
                e'8
                c'8
            }

        ..  figure:: ../_images/Phaser-x1gn7mv4jq.png

        >>> container = abjad.Container(r"\time 3/8 c'8 d'8 e'8")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        forward_bias=0.0,
        ...                        )
        >>> notes = phaser.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'8
                d'8
                e'8
                e'16
                c'8
                d'8
                e'16
                e'8
                c'8
                d'8
            }

        ..  figure:: ../_images/Phaser-g6y94qiwfro.png

    :attr:`forward_bias` between ``0.0`` and ``1.0``:
        Setingt :attr:`forward_bias` to a value in between ``0.0`` and ``1.0``
        will result in random steps being taken forward or backward, according
        to the bias.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 4),
        ...                        forward_bias=0.5,
        ...                        )
        >>> notes = phaser.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
                d'4
                e'4
                f'4
                c'4
                c'4
                d'4
                e'4
                f'4
                f'4
                c'4
                d'4
                e'4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Phaser-lu2p8o1g6p.png

    :attr:`max_steps`:
        Setting the keyword argument :attr:`max_steps` to a value larger than
        ``1`` will result in a random number of steps (between ``1`` and
        :attr:`max_steps`) being applied at each call.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 8),
        ...                        max_steps=4,
        ...                        )
        >>> notes = phaser.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                c''8
                f'8
                g'8
                a'8
                b'8
                c''8
                c'8
                d'8
                e'8
                c''8
                c'8
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                f'8
                g'8
                a'8
                b'8
                c''8
                c'8
                d'8
                e'8
                a'8
                b'8
                c''8
                c'8
                d'8
                e'8
                f'8
                g'8
            }

        ..  figure:: ../_images/Phaser-pvbugjwu5qd.png

    :func:`len()`:
        The function :func:`len()` can be used to get the total number of steps
        that are necessary to return to the initial container.

        >>> container = abjad.Container(r"c'1")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        )
        >>> len(phaser)
        16
        >>> container = abjad.Container(r"c'1")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 4),
        ...                        )
        >>> len(phaser)
        4
        >>> container = abjad.Container(r"\time 3/4 c'2.")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 2),
        ...                        )
        >>> len(phaser)
        3

    :meth:`output_all`:
        To run through the whole process and output it as a single container,
        from the initial head position until the process outputs the single
        last element, use the method :meth:`output_all`.

        >>> container = abjad.Container(r"\time 3/4 c'4. d'4.")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 4),
        ...                        )
        >>> notes = phaser.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4.
                d'4.
                c'8
                d'4.
                c'4
                d'4
                c'4.
                d'8
                c'4.
                d'4.
            }

        ..  figure:: ../_images/Phaser-y6r36v26mk7.png

    ``cycle_back_to_first``:
        By default, :meth:`output_all` will cycle back to the very first
        window. To stop at the iteration step just before looping back to the
        initial container, set the keyword argument ``cycle_back_to_first`` to
        ``False``.

        >>> container = abjad.Container(r"\time 3/4 c'4. d'4.")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 4),
        ...                        )
        >>> notes = phaser.output_all(cycle_back_to_first=False)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4.
                d'4.
                c'8
                d'4.
                c'4
                d'4
                c'4.
                d'8
            }

        ..  figure:: ../_images/Phaser-1lc6vkz1wuf.png

    :meth:`output_n`:
        To run through just part of the process and output it as a single
        container, starting from the initial head position, use the method
        :meth:`output_n` and pass the number of iterations as argument.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 32),
        ...                        )
        >>> notes = phaser.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
                c'8..
                d'32
                ~
                d'8..
                e'32
                ~
                e'8..
                f'32
                ~
                f'8..
                c'32
                c'8.
                d'16
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
                ~
                f'8.
                c'16
            }

        ..  figure:: ../_images/Phaser-ii58nvyi3c.png

    ``tie_identical_pitches``:
        Both :meth:`output_n` and :meth:`output_all` methods can receive the
        keyword argument ``tie_identical_pitches``, which will tie identical
        pitches across iterations of the process when set to ``True`` (default
        is ``False``). Compare the output below with the example just above.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 32),
        ...                        )
        >>> notes = phaser.output_n(3, tie_identical_pitches=True)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
                c'8..
                d'32
                ~
                d'8..
                e'32
                ~
                e'8..
                f'32
                ~
                f'8..
                c'32
                ~
                c'8.
                d'16
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
                ~
                f'8.
                c'16
            }

        ..  figure:: ../_images/Phaser-gvzlwz9gol6.png

    :attr:`remove_unterminated_ties`:
        All methods that call the phasing process (:meth:`__call__`,
        :meth:`__next__`, :meth:`output_all`, :meth:`output_n`) remove
        unterminated ties at the end of a selection, which are a result of a
        logical tie being split at that point. Use the optional keyword
        argument ``remove_unterminated_ties=False`` when initialising the
        phaser to disable this behaviour.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        process_on_first_call=True,
        ...                        remove_unterminated_ties=False,
        ...                        )
        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
                ~
                f'8.
                c'16
                ~
            }

        ..  figure:: ../_images/Phaser-hg27xwbz4t.png

    Time signature changes:
        This class handles time signature changes.

        >>> container = abjad.Container(
        ...     r"\time 2/4 c'2 \time 3/8 d'4. \time 2/4 e'2"
        ... )
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 8),
        ...                        )
        >>> notes = phaser.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'2
                \time 3/8
                d'4.
                \time 2/4
                e'2
                c'4.
                d'8
                ~
                \time 3/8
                d'4
                e'8
                ~
                \time 2/4
                e'4.
                c'8
                c'4
                d'4
                ~
                \time 3/8
                d'8
                e'4
                ~
                \time 2/4
                e'4
                c'4
            }

        ..  figure:: ../_images/Phaser-vw3gfizbkg.png

    Indicators:
        This class can handle dynamics and articulations too. When a logical
        tie is split into two during the phasing process, dynamics and
        articulations are passed on to both of them.

        >>> container = abjad.Container(r"c'4-.\p\< d'4--\f e'4->\p f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 8),
        ...                        )
        >>> staff = abjad.Staff()
        >>> notes = phaser.output_n(5)
        >>> staff.append(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                - \staccato
                \<
                d'4
                \f
                - \tenuto
                e'4
                \p
                - \accent
                f'4
                c'8
                - \staccato
                \<
                d'4
                \f
                - \tenuto
                e'8
                \p
                - \accent
                ~
                e'8
                f'4
                c'8
                - \staccato
                d'4
                \f
                - \tenuto
                e'4
                \p
                - \accent
                f'4
                c'4
                - \staccato
                d'8
                \f
                - \tenuto
                e'4
                \p
                - \accent
                f'8
                ~
                f'8
                c'4
                - \staccato
                d'8
                \f
                - \tenuto
                e'4
                \p
                - \accent
                f'4
                c'4
                - \staccato
                \<
                d'4
                \f
                - \tenuto
            }

        ..  figure:: ../_images/Phaser-978yobu5f2q.png

    Slurs and hairpins:
        Slurs and hairpins are also supported.

        >>> container = abjad.Container(
        ...     r"c'2(\p\< d'4. e'8\f f'4\p\> g'2 a'4\pp)"
        ... )
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        )
        >>> notes = phaser.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'2
                \p
                \<
                (
                d'4.
                e'8
                \f
                f'4
                \p
                \>
                g'2
                a'4
                \pp
                )
                c'4..
                \p
                \<
                (
                d'16
                ~
                d'4
                ~
                d'16
                e'8
                \f
                f'16
                \p
                ~
                f'8.
                \>
                g'16
                ~
                g'4
                ~
                g'8.
                a'16
                \pp
                ~
                a'8.
                )
                c'16
                \p
                (
                c'4.
                \<
                d'8
                ~
                d'4
                e'8
                \f
                f'8
                \p
                ~
                f'8
                \>
                g'8
                ~
                g'4
                ~
                g'8
                a'4
                \pp
                )
                c'8
                \p
                (
                c'4
                ~
                c'16
                \<
                d'8.
                ~
                d'8.
                e'16
                \f
                ~
                e'16
                f'8.
                \p
                ~
                f'16
                \>
                g'8.
                ~
                g'4
                ~
                g'16
                a'8.
                \pp
                ~
                a'16
                )
                c'8.
                \p
                (
                c'4
                \<
                d'4
                ~
                d'8
                e'8
                \f
                f'4
                \p
                \>
                g'2
                a'4
                \pp
                )
                c'4
                \p
            }

        ..  figure:: ../_images/Phaser-pxj7axmh58e.png

    ..  tip::

        The functions |auxjad.mutate.remove_repeated_dynamics()| and
        |auxjad.mutate.reposition_clefs()| can be used to clean the output
        and remove repeated dynamics and unnecessary clef changes.

    ..  warning::

        Do note that some elements that span multiple notes (such as ottava
        indicators, manual beams, etc.) can become problematic when notes
        containing them are split into two. As a rule of thumb, it is always
        better to attach those to the music after the fading process has ended.

    :attr:`contents`:
        Use the :attr:`contents` property to read as well as overwrite the
        contents of the phaser. Notice that the phasing process will start from
        the beginning of the new container.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        )
        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Phaser-ns0dbcanp5.png

        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                d'16
                ~
                d'8.
                e'16
                ~
                e'8.
                f'16
                ~
                f'8.
                c'16
            }

        ..  figure:: ../_images/Phaser-utcxb4w611s.png

        >>> phaser.contents = abjad.Container(r"c'16 d'16 e'16 f'16 g'2.")
        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'16
                d'16
                e'16
                f'16
                g'2.
            }

        ..  figure:: ../_images/Phaser-w0whe6c7jp.png

        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                d'16
                e'16
                f'16
                g'16
                ~
                g'2
                ~
                g'8.
                c'16
            }

        ..  figure:: ../_images/Phaser-cmo73lo14wo.png

    Tweaking |abjad.Meter.rewrite_meter()|:
        This function uses the default logical tie splitting algorithm from
        |abjad.Meter.rewrite_meter()|.

        >>> container = abjad.Container(r"c'4. d'8 e'2")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        )
        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4.
                d'8
                e'2
            }

        ..  figure:: ../_images/Phaser-uhfkf8mzu4i.png

        Set :attr:`boundary_depth` to a different number to change its
        behaviour.

        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        boundary_depth=1,
        ...                        )
        >>> notes = phaser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                ~
                c'8
                d'8
                e'2
            }

        ..  figure:: ../_images/Phaser-l6pvwhsj1hh.png

        Other arguments available for tweaking the output of
        |abjad.Meter.rewrite_meter()| are :attr:`maximum_dot_count` and
        :attr:`rewrite_tuplets`, which work exactly as the identically named
        arguments of |abjad.Meter.rewrite_meter()|.

        This class also accepts the arguments ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, ``fuse_triple_meter``, and
        ``extract_trivial_tuplets``, which are passed on to
        |auxjad.mutate.prettify_rewrite_meter()| (the latter can be disabled
        by setting ``prettify_rewrite_meter`` to ``False``). See the
        documentation of this function for more details on these arguments.

    :attr:`omit_time_signatures`:
        To disable time signatures altogether, initialise this class with the
        keyword argument :attr:`omit_time_signatures` set to ``True`` (default
        is ``False``), or use the :attr:`omit_time_signatures` property after
        initialisation.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 8),
        ...                        omit_time_signatures=True,
        ...                        )
        >>> notes = phaser.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                c'8
                d'4
                e'4
                c'8
                d'4
                e'4
                c'4
            }

        ..  figure:: ../_images/Phaser-1rdkecll151.png

    ..  tip::

        All methods that return an |abjad.Selection| will add an initial time
        signature to it. The :meth:`output_n` and :meth:`output_all` methods
        automatically remove repeated time signatures. When joining selections
        output by multiple method calls, use
        |auxjad.mutate.remove_repeated_time_signatures()| on the whole
        container after fusing the selections to remove any unecessary time
        signature changes.

    ..  warning::

        This class can handle tuplets, but the output is often quite complex.
        Although the result will be rhythmically correct, consecutive tuplets
        are not fused together, and tuplets may be output off-beat. This
        functionality should be considered experimental.

        >>> container = abjad.Container(r"\times 2/3 {c'8 d'8 e'8} d'2.")
        >>> phaser = auxjad.Phaser(container,
        ...                        step_size=(1, 16),
        ...                        )
        >>> notes = phaser.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    \time 4/4
                    c'8
                    d'8
                    e'8
                }
                d'2.
                \times 2/3
                {
                    c'32
                    d'16
                    ~
                    d'16
                    e'8
                }
                d'16
                ~
                d'2
                ~
                d'8.
                c'16
                \times 2/3
                {
                    d'16
                    e'8
                }
                d'8
                ~
                d'2
                ~
                d'8
                \times 2/3
                {
                    c'8
                    d'16
                }
            }

        ..  figure:: ../_images/Phaser-lqvc5p5i25b.png
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_contents",
        "_pivot_point",
        "_step_size",
        "_max_steps",
        "_forward_bias",
        "_remove_unterminated_ties",
        "_current_window",
        "_is_first_window",
        "_contents_length",
        "_omit_time_signatures",
        "_boundary_depth",
        "_maximum_dot_count",
        "_rewrite_tuplets",
        "_process_on_first_call",
        "_prettify_rewrite_meter",
        "_extract_trivial_tuplets",
        "_fuse_across_groups_of_beats",
        "_fuse_quadruple_meter",
        "_fuse_triple_meter",
    )

    ### INITIALISER ###

    def __init__(
        self,
        contents: abjad.Container,
        *,
        step_size: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
        ],
        max_steps: int = 1,
        forward_bias: float = 1.0,
        process_on_first_call: bool = False,
        remove_unterminated_ties: bool = True,
        omit_time_signatures: bool = False,
        boundary_depth: Optional[int] = None,
        maximum_dot_count: Optional[int] = None,
        rewrite_tuplets: bool = True,
        prettify_rewrite_meter: bool = True,
        extract_trivial_tuplets: bool = True,
        fuse_across_groups_of_beats: bool = True,
        fuse_quadruple_meter: bool = True,
        fuse_triple_meter: bool = True,
    ) -> None:
        self.contents = contents
        self._pivot_point = abjad.Duration(0)
        self.step_size = step_size
        self.max_steps = max_steps
        self.forward_bias = forward_bias
        self.remove_unterminated_ties = remove_unterminated_ties
        self.omit_time_signatures = omit_time_signatures
        self.boundary_depth = boundary_depth
        self.maximum_dot_count = maximum_dot_count
        self.rewrite_tuplets = rewrite_tuplets
        self.prettify_rewrite_meter = prettify_rewrite_meter
        self.extract_trivial_tuplets = extract_trivial_tuplets
        self.fuse_across_groups_of_beats = fuse_across_groups_of_beats
        self.fuse_quadruple_meter = fuse_quadruple_meter
        self.fuse_triple_meter = fuse_triple_meter
        self.process_on_first_call = process_on_first_call
        self._is_first_window = True

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return abjad.lilypond(self._contents)

    def __len__(self) -> int:
        r"""Returns the length of :attr:`contents` in terms of
        :attr:`step_size`.
        """
        proportion = self._contents_length / self._step_size
        return int(proportion * proportion.denominator)

    def __call__(self) -> abjad.Selection:
        r"""Calls the phaser process for one iteration, returning an
        |abjad.Selection|.
        """
        if not self._is_first_window or self._process_on_first_call:
            self._move_pivot_point()
        self._make_music()
        self._is_first_window = False
        return self.current_window

    def __next__(self) -> abjad.Selection:
        r"""Calls the phaser process for one iteration, returning an
        |abjad.Selection|.
        """
        if not self._is_first_window or self._process_on_first_call:
            self._move_pivot_point()
            if self._done:
                raise StopIteration
        self._make_music()
        self._is_first_window = False
        return self.current_window

    def __iter__(self) -> None:
        r"""Returns an iterator, allowing instances to be used as iterators."""
        return self

    ### PUBLIC METHODS ###

    def output_all(
        self,
        *,
        cycle_back_to_first: bool = True,
        tie_identical_pitches: bool = False,
    ) -> abjad.Selection:
        r"""Goes through the whole phasing process and outputs a single
        |abjad.Selection|.
        """
        if not isinstance(cycle_back_to_first, bool):
            raise TypeError("'cycle_back_to_first' must be 'bool'")
        if not isinstance(tie_identical_pitches, bool):
            raise TypeError("'tie_identical_pitches' must be 'bool'")
        dummy_container = abjad.Container()
        while True:
            selection = self.__call__()
            if not self._done:
                if tie_identical_pitches:
                    self._tie_identical_pitches(selection, dummy_container)
                dummy_container.append(selection)
            else:
                break
        if cycle_back_to_first:
            if tie_identical_pitches:
                self._tie_identical_pitches(selection, dummy_container)
            dummy_container.append(selection)
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.reposition_dynamics(dummy_container[:])
        mutate.reposition_slurs(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    def output_n(
        self,
        n: int,
        *,
        tie_identical_pitches: bool = False,
    ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the phasing process and outputs a
        single |abjad.Selection|.
        """
        if not isinstance(n, int):
            raise TypeError("first positional argument must be 'int'")
        if n < 1:
            raise ValueError("first positional argument must be a positive 'int'")
        if not isinstance(tie_identical_pitches, bool):
            raise TypeError("'tie_identical_pitches' must be 'bool'")
        dummy_container = abjad.Container()
        for _ in range(n):
            selection = self.__call__()
            if tie_identical_pitches:
                self._tie_identical_pitches(selection, dummy_container)
            dummy_container.append(selection)
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.reposition_dynamics(dummy_container[:])
        mutate.reposition_slurs(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    ### PRIVATE METHODS ###

    def _move_pivot_point(self) -> None:
        r"""Moves the pivot point by a certain number of steps of fixed size,
        either forwards or backwards according to the forward bias.
        """
        step = self._step_size * random.randint(1, self._max_steps)
        diretion = self._biased_choice(self._forward_bias)
        self._pivot_point += step * diretion

    def _make_music(self) -> None:
        r"""Applies the phasing process and handles the output container."""
        dummy_container = self._phase_contents()
        # dealing with dynamics
        mutate.reposition_dynamics(dummy_container[:])
        # adding time signatures back and rewriting meter
        time_signatures = get.time_signature_list(
            self._contents,
            do_not_use_none=True,
        )
        mutate.enforce_time_signature(
            dummy_container,
            time_signatures,
            boundary_depth=self._boundary_depth,
            maximum_dot_count=self._maximum_dot_count,
            rewrite_tuplets=self._rewrite_tuplets,
            prettify_rewrite_meter=self._prettify_rewrite_meter,
            extract_trivial_tuplets=self._extract_trivial_tuplets,
            fuse_across_groups_of_beats=self._fuse_across_groups_of_beats,
            fuse_quadruple_meter=self._fuse_quadruple_meter,
            fuse_triple_meter=self._fuse_triple_meter,
        )
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    def _phase_contents(self) -> abjad.Container:
        r"""This method phases :attr:`contents` using ``_pivot_point`` as the
        pivot point.
        """
        dummy_container = abjad.mutate.copy(self._contents)
        pivot = self._pivot_point % self._contents_length
        # splitting leaves at both slicing points
        if pivot > abjad.Duration(0):
            abjad.mutate.split(dummy_container[:], [pivot])
            # finding start and end indeces for the window
            for start in range(len(dummy_container)):
                if abjad.get.duration(dummy_container[: start + 1]) > pivot:
                    break
            last_leaf = dummy_container[:start].leaf(-1)
            # copying indicators to both leaves
            indicators_tuple = (
                abjad.Articulation,
                abjad.Clef,
                abjad.Dynamic,
                abjad.Fermata,
                abjad.KeySignature,
                abjad.LilyPondLiteral,
                abjad.MetronomeMark,
                abjad.Ottava,
                abjad.StaffChange,
                abjad.StartHairpin,
                abjad.StartPhrasingSlur,
                abjad.StartSlur,
                abjad.StopHairpin,
                abjad.StopPhrasingSlur,
                abjad.StopSlur,
            )
            if abjad.get.indicator(last_leaf, abjad.Tie):  # i.e. split
                for indicator in abjad.get.indicators(last_leaf):
                    if isinstance(indicator, indicators_tuple):
                        first_leaf = dummy_container[start:].leaf(0)
                        abjad.attach(indicator, first_leaf)
            # removing ties of splitted logical tie if necessary
            if self._remove_unterminated_ties and abjad.get.indicator(last_leaf, abjad.Tie):
                abjad.detach(abjad.Tie, last_leaf)
            # joining two subcontainers
            dummy_end_container = abjad.Container(abjad.mutate.copy(dummy_container[:start]))
            dummy_container = abjad.Container(abjad.mutate.copy(dummy_container[start:]))
            dummy_container.extend(dummy_end_container)
            dummy_end_container[:] = []
        return dummy_container

    def _get_lilypond_format(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return self.__repr__()

    @staticmethod
    def _tie_identical_pitches(
        currrent_selection: abjad.Selection,
        previous_container: abjad.Container,
    ) -> None:
        r"""Ties identical pitches when joining windows."""
        if len(previous_container) == 0:
            return
        first_leaf = currrent_selection.leaf(0)
        last_leaf = abjad.select(previous_container).leaf(-1)
        if get.leaves_are_tieable((first_leaf, last_leaf)) and not abjad.get.indicator(
            last_leaf, abjad.Tie
        ):
            abjad.attach(abjad.Tie(), last_leaf)

    @staticmethod
    def _biased_choice(
        forward_bias: float,
    ) -> None:
        r"""Returns either +1 or -1 according to a forward bias value."""
        weights = [forward_bias, 1.0 - forward_bias]
        return random.choices([1, -1], weights=weights)[0]

    @staticmethod
    def _remove_all_time_signatures(
        container: abjad.Container,
    ) -> None:
        r"""Removes all time signatures of an |abjad.Container|."""
        for leaf in abjad.select(container).leaves():
            if abjad.get.effective(leaf, abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r"""The |abjad.Container| to be phased."""
        return abjad.mutate.copy(self._contents)

    @contents.setter
    def contents(
        self,
        contents: abjad.Container,
    ) -> None:
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child class")
        if not abjad.select(contents).leaves().are_contiguous_logical_voice():
            raise ValueError("'contents' must be contiguous logical voice")
        if isinstance(contents, abjad.Score):
            self._contents = abjad.mutate.copy(contents[0])
        elif isinstance(contents, abjad.Tuplet):
            self._contents = abjad.Container([abjad.mutate.copy(contents)])
        else:
            self._contents = abjad.mutate.copy(contents)
        dummy_container = abjad.mutate.copy(contents)
        self._current_window = dummy_container[:]
        dummy_container[:] = []
        self._contents_length = abjad.get.duration(self._contents[:])
        self._pivot_point = abjad.Duration(0)
        self._is_first_window = True

    @property
    def current_window(self) -> abjad.Selection:
        r"""Read-only property, returns the previously output selection."""
        current_window = abjad.mutate.copy(self._current_window)
        if self._omit_time_signatures:
            self._remove_all_time_signatures(current_window)
        return current_window

    @property
    def pivot_point(self) -> abjad.Duration:
        r"""Read-only property, returns the position of the pivot point."""
        return self._pivot_point

    @property
    def step_size(self) -> abjad.Duration:
        r"""The size of each step when moving the pivot point."""
        return self._step_size

    @step_size.setter
    def step_size(
        self,
        step_size: Union[
            int,
            float,
            str,
            tuple[int],
            abjad.Duration,
        ],
    ) -> None:
        if not isinstance(step_size, (abjad.Duration, str, tuple, int, float)):
            raise TypeError("'step_size' must be 'abjad.Duration', 'str', 'tuple', or a number")
        self._step_size = abjad.Duration(step_size)

    @property
    def max_steps(self) -> int:
        r"""The maximum number of steps per operation."""
        return self._max_steps

    @max_steps.setter
    def max_steps(
        self,
        max_steps: int,
    ) -> None:
        if not isinstance(max_steps, int):
            raise TypeError("'max_steps' must be 'int'")
        if max_steps < 1:
            raise ValueError("'max_steps' must be greater than zero")
        self._max_steps = max_steps

    @property
    def forward_bias(self) -> float:
        r"""The chance of the window moving forward instead of backwards. It
        should range from 0.0 to 1.0 (default 1.0, which means the window can
        only move forwards. A value of 0.5 gives 50% chance of moving forwards
        while a value of 0.0 will move the window only backwards).
        """
        return self._forward_bias

    @forward_bias.setter
    def forward_bias(
        self,
        forward_bias: float,
    ) -> None:
        if not isinstance(forward_bias, float):
            raise TypeError("'forward_bias' must be 'float'")
        if forward_bias < 0.0 or forward_bias > 1.0:
            raise ValueError("'forward_bias' must be between 0.0 and 1.0")
        self._forward_bias = forward_bias

    @property
    def omit_time_signatures(self) -> bool:
        r"""When ``True``, all time signatures will be omitted from the
        output.
        """
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(
        self,
        omit_time_signatures: bool,
    ) -> None:
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def remove_unterminated_ties(self) -> bool:
        r"""When ``True``, the last element of the |abjad.Selection| returned
        by a call will have any ties removed. This means that splitted logical
        ties will not tie accross multiple calls.
        """
        return self._remove_unterminated_ties

    @remove_unterminated_ties.setter
    def remove_unterminated_ties(
        self,
        remove_unterminated_ties: bool,
    ) -> None:
        if not isinstance(remove_unterminated_ties, bool):
            raise TypeError("'remove_unterminated_ties' must be 'bool'")
        self._remove_unterminated_ties = remove_unterminated_ties

    @property
    def boundary_depth(self) -> Union[int, None]:
        r"""Sets the argument ``boundary_depth`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._boundary_depth

    @boundary_depth.setter
    def boundary_depth(
        self,
        boundary_depth: Optional[int],
    ) -> None:
        if boundary_depth is not None:
            if not isinstance(boundary_depth, int):
                raise TypeError("'boundary_depth' must be 'int'")
        self._boundary_depth = boundary_depth

    @property
    def maximum_dot_count(self) -> Union[int, None]:
        r"""Sets the argument ``maximum_dot_count`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._maximum_dot_count

    @maximum_dot_count.setter
    def maximum_dot_count(
        self,
        maximum_dot_count: Optional[int],
    ) -> None:
        if maximum_dot_count is not None:
            if not isinstance(maximum_dot_count, int):
                raise TypeError("'maximum_dot_count' must be 'int'")
        self._maximum_dot_count = maximum_dot_count

    @property
    def rewrite_tuplets(self) -> bool:
        r"""Sets the argument ``rewrite_tuplets`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._rewrite_tuplets

    @rewrite_tuplets.setter
    def rewrite_tuplets(
        self,
        rewrite_tuplets: bool,
    ) -> None:
        if not isinstance(rewrite_tuplets, bool):
            raise TypeError("'rewrite_tuplets' must be 'bool'")
        self._rewrite_tuplets = rewrite_tuplets

    @property
    def prettify_rewrite_meter(self) -> bool:
        r"""Used to enable or disable the mutation
        |auxjad.mutate.prettify_rewrite_meter()| (default ``True``).
        """
        return self._prettify_rewrite_meter

    @prettify_rewrite_meter.setter
    def prettify_rewrite_meter(
        self,
        prettify_rewrite_meter: bool,
    ) -> None:
        if not isinstance(prettify_rewrite_meter, bool):
            raise TypeError("'prettify_rewrite_meter' must be 'bool'")
        self._prettify_rewrite_meter = prettify_rewrite_meter

    @property
    def extract_trivial_tuplets(self) -> bool:
        r"""Sets the argument ``extract_trivial_tuplets`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._extract_trivial_tuplets

    @extract_trivial_tuplets.setter
    def extract_trivial_tuplets(
        self,
        extract_trivial_tuplets: bool,
    ) -> None:
        if not isinstance(extract_trivial_tuplets, bool):
            raise TypeError("'extract_trivial_tuplets' must be 'bool'")
        self._extract_trivial_tuplets = extract_trivial_tuplets

    @property
    def fuse_across_groups_of_beats(self) -> bool:
        r"""Sets the argument ``fuse_across_groups_of_beats`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_across_groups_of_beats

    @fuse_across_groups_of_beats.setter
    def fuse_across_groups_of_beats(
        self,
        fuse_across_groups_of_beats: bool,
    ) -> None:
        if not isinstance(fuse_across_groups_of_beats, bool):
            raise TypeError("'fuse_across_groups_of_beats' must be 'bool'")
        self._fuse_across_groups_of_beats = fuse_across_groups_of_beats

    @property
    def fuse_quadruple_meter(self) -> bool:
        r"""Sets the argument ``fuse_quadruple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_quadruple_meter

    @fuse_quadruple_meter.setter
    def fuse_quadruple_meter(
        self,
        fuse_quadruple_meter: bool,
    ) -> None:
        if not isinstance(fuse_quadruple_meter, bool):
            raise TypeError("'fuse_quadruple_meter' must be 'bool'")
        self._fuse_quadruple_meter = fuse_quadruple_meter

    @property
    def fuse_triple_meter(self) -> bool:
        r"""Sets the argument ``fuse_triple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_triple_meter

    @fuse_triple_meter.setter
    def fuse_triple_meter(
        self,
        fuse_triple_meter: bool,
    ) -> None:
        if not isinstance(fuse_triple_meter, bool):
            raise TypeError("'fuse_triple_meter' must be 'bool'")
        self._fuse_triple_meter = fuse_triple_meter

    @property
    def process_on_first_call(self) -> bool:
        r"""If ``True`` then :attr:`contents` will be processed in the very
        first call.
        """
        return self._process_on_first_call

    @process_on_first_call.setter
    def process_on_first_call(
        self,
        process_on_first_call: bool,
    ) -> None:
        if not isinstance(process_on_first_call, bool):
            raise TypeError("'process_on_first_call' must be 'bool'")
        self._process_on_first_call = process_on_first_call

    ### PRIVATE PROPERTIES ###

    @property
    def _done(self) -> bool:
        r""":obj:`bool` indicating whether the process is done (i.e. whether
        the pivot point has overtaken the :attr:`contents`'s length). Only
        :meth:`__next__` and :meth:`output_all` make use of it, since regular
        calls make use of the module of the position of the pivot point in
        relation to the duration of :attr:`contents`, allowing for infinitely
        many calls.
        """
        return (
            self._pivot_point >= self._contents_length
            or self._pivot_point <= -self._contents_length
        )



import abjad

from .. import get, mutate


class Repeater:
    r"""Takes an |abjad.Container| (or child class) as input and outputs an
    |abjad.Selection| with ``n`` repetitions. It can be of type unfold or
    volta.

    Basic usage:
        Calling the object will return an |abjad.Selection| generated by the
        repeating process, which is of type unfold by default (more on this
        below). The argument of :meth:`__call__()` defines the number of
        repetitions.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Repeater-aK6JHsJPDM.png

        >>> repeater = auxjad.Repeater(container)
        >>> notes = repeater(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Repeater-tigd5dwtszh.png

        The property :attr:`current_window` can be used to access the last
        results.

        >>> notes = repeater.current_window()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Repeater-hg86wd75fvp.png

    :attr:`repeat_type`:
        Use the :attr:`repeat_type` property to set the type of the repeater.
        It takes a single string, which should be either ``'unfold'`` or
        ``'volta'``. Unfold is the default mode i.e. the repeater outputs a
        selection of ``n`` identical consecutive measures. In volta mode,
        repeat bars are added as well as a written indication of the number of
        repeats. Compare:

        >>> container = abjad.Container(r"c'2 d'2")
        >>> repeater = auxjad.Repeater(container,
        ...                            repeat_type='unfold',
        ...                            )
        >>> notes = repeater(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2
                d'2
                c'2
                d'2
                c'2
                d'2
                c'2
                d'2
                c'2
                d'2
            }

        ..  figure:: ../_images/Repeater-vQi9k3oMox.png

        >>> container = abjad.Container(r"c'2 d'2")
        >>> repeater = auxjad.Repeater(container,
        ...                            repeat_type='volta',
        ...                            )
        >>> notes = repeater(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \repeat volta 5
                {
                    c'2
                    d'2
                    \tweak RehearsalMark.self-alignment-X #RIGHT
                    \tweak RehearsalMark.break-visibility
                        #begin-of-line-invisible
                    \mark \markup{\box "5×"}
                }
            }

        ..  figure:: ../_images/Repeater-k8qwVVoMu9.png

    Typical usage of volta mode:
        The volta mode can be used to construct containers in the following
        way:

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> repeater = auxjad.Repeater(container,
        ...                            repeat_type='volta',
        ...                            )
        >>> notes = repeater(3)
        >>> staff = abjad.Staff(notes)
        >>> repeater.contents = abjad.Container(r"g'2 a'2")
        >>> notes = repeater(2)
        >>> staff.append(notes)
        >>> repeater.contents = abjad.Container(r"b'16 c''16 d''16 e''16 r2.")
        >>> notes = repeater(5)
        >>> staff.append(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \repeat volta 3
                {
                    c'4
                    d'4
                    e'4
                    f'4
                    \tweak RehearsalMark.self-alignment-X #RIGHT
                    \tweak RehearsalMark.break-visibility
                        #begin-of-line-invisible
                    \mark \markup{\box "3×"}
                }
                \repeat volta 2
                {
                    g'2
                    a'2
                    \tweak RehearsalMark.self-alignment-X #RIGHT
                    \tweak RehearsalMark.break-visibility
                        #begin-of-line-invisible
                    \mark \markup{\box "2×"}
                }
                \repeat volta 5
                {
                    b'16
                    c''16
                    d''16
                    e''16
                    r2.
                    \tweak RehearsalMark.self-alignment-X #RIGHT
                    \tweak RehearsalMark.break-visibility
                        #begin-of-line-invisible
                    \mark \markup{\box "5×"}
                }
            }

        ..  figure:: ../_images/Repeater-KbjMitAIs5.png

    :attr:`include_2x_volta_text`:
        By default, the written indication ``nx`` is added to all repeats in
        volta mode. To omit it when ``n`` is ``2``, set
        :attr:`include_2x_volta_text` to ``False``:

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> repeater = auxjad.Repeater(container,
        ...                            repeat_type='volta',
        ...                            include_2x_volta_text=False,
        ...                            )
        >>> notes = repeater(3)
        >>> staff = abjad.Staff(notes)
        >>> repeater.contents = abjad.Container(r"g'2 a'2")
        >>> notes = repeater(2)
        >>> staff.append(notes)
        >>> repeater.contents = abjad.Container(r"b'16 c''16 d''16 e''16 r2.")
        >>> notes = repeater(5)
        >>> staff.append(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \repeat volta 3
                {
                    c'4
                    d'4
                    e'4
                    f'4
                    \tweak RehearsalMark.self-alignment-X #RIGHT
                    \tweak RehearsalMark.break-visibility
                        #begin-of-line-invisible
                    \mark \markup{\box "3×"}
                }
                \repeat volta 2
                {
                    g'2
                    a'2
                }
                \repeat volta 5
                {
                    b'16
                    c''16
                    d''16
                    e''16
                    r2.
                    \tweak RehearsalMark.self-alignment-X #RIGHT
                    \tweak RehearsalMark.break-visibility
                        #begin-of-line-invisible
                    \mark \markup{\box "5×"}
                }
            }

        ..  figure:: ../_images/Repeater-vW77Zi6NZJ.png

    Time signatures:
        This class handles different time signatures.

        >>> container = abjad.Container(r"\time 3/4 c'2. \time 2/4 r2 g'2")
        >>> repeater = auxjad.Repeater(container)
        >>> notes = repeater(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                \time 2/4
                r2
                g'2
                \time 3/4
                c'2.
                \time 2/4
                r2
                g'2
                \time 3/4
                c'2.
                \time 2/4
                r2
                g'2
            }

        ..  figure:: ../_images/Repeater-fqkjxhegzmv.png

    Underfull containers:
        Containers that are not fully filled in are automatically closed by
        this class in its output. Containers without a time signature are
        assumed to be in ``4/4`` (which is LilyPond's default).

        >>> container = abjad.Container(r"c'4 d'4 e'4")
        >>> repeater = auxjad.Repeater(container)
        >>> notes = repeater(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                c'4
                d'4
                e'4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/Repeater-k4hxxghalwh.png

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 f'2")
        >>> repeater = auxjad.Repeater(container)
        >>> notes = repeater(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                \time 2/4
                f'2
                \time 3/4
                c'4
                d'4
                e'4
                \time 2/4
                f'2
            }

        ..  figure:: ../_images/Repeater-fee3qe1vdjl.png

    Using as iterator:
        The instances of this class can also be used as an iterator, which can
        then be used in a for loop. Note that unlike the methods
        :meth:`__call__()` and :meth:`output_n`, time signatures are added to
        each window returned by the shuffler. Use the function
        |auxjad.mutate.remove_repeated_time_signatures()| to clean the output
        when using this class in this way. It is also important to note that a
        ``break`` statement is needed when using this class as an iterator. The
        reason is that repeating is a process that can happen indefinitely
        (unlike some of the other classes in this library).

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> repeater = auxjad.Repeater(container)
        >>> staff = abjad.Staff()
        >>> for window in repeater:
        ...     staff.append(window)
        ...     if abjad.get.duration(staff) == abjad.Duration((9, 4)):
        ...         break
        >>> auxjad.mutate.remove_repeated_time_signatures(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'4
                d'4
                e'4
                c'4
                d'4
                e'4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/Repeater-8oouugbk5zc.png

    Arguments and properties:
        This class can take many optional keyword arguments during its
        creation, besides attr:`repeat_type` and attr:`include_2x_volta_text`.
        attr:`omit_time_signatures` will remove all time signatures from the
        output while :attr:`force_identical_time_signatures` will force all
        time signatures (including repeated ones) to be added to the output
        (both are ``False`` by default). When set to ``True``, the properties
        :attr:`reposition_clefs`, :attr:`reposition_dynamics`, and
        :attr:`reposition_slurs` will invoke the mutations
        |auxjad.mutate.reposition_clefs()|,
        |auxjad.mutate.reposition_dynamics()|, and
        |auxjad.mutate.reposition_slurs()| (default values are ``True``).
        Check their documentation for more information on how they operate.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> repeater = auxjad.Repeater(container,
        ...                            repeat_type='volta',
        ...                            include_2x_volta_text=False,
        ...                            omit_time_signatures=False,
        ...                            force_identical_time_signatures=False,
        ...                            reposition_clefs=True,
        ...                            reposition_dynamics=True,
        ...                            reposition_slurs=True,
        ...                            )
        >>> repeater.repeat_type
        'volta'
        >>> repeater.include_2x_volta_text
        False
        >>> repeater.omit_time_signatures
        False
        >>> repeater.force_identical_time_signatures
        False
        >>> repeater.reposition_clefs
        True
        >>> repeater.reposition_dynamics
        True
        >>> repeater.reposition_slurs
        True

        Use the properties below to change these values after initialisation.

        >>> repeater.repeat_type = 'unfold'
        >>> repeater.include_2x_volta_text = True
        >>> repeater.omit_time_signatures = True
        >>> repeater.force_identical_time_signatures = True
        >>> repeater.reposition_clefs = False
        >>> repeater.reposition_dynamics = False
        >>> repeater.reposition_slurs = False
        >>> repeater.repeat_type
        'unfold'
        >>> repeater.include_2x_volta_text
        True
        >>> repeater.omit_time_signatures
        True
        >>> repeater.force_identical_time_signatures
        True
        >>> repeater.reposition_clefs
        False
        >>> repeater.reposition_dynamics
        False
        >>> repeater.reposition_slurs
        False

    :attr:`contents`:
        Use the :attr:`contents` property to read as well as overwrite the
        contents of the repeater.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> repeater = auxjad.Repeater(container)
        >>> notes = repeater(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Repeater-f1kqq128afw.png

        >>> repeater.contents = abjad.Container(r"c'16 d'16 e'16 f'16 g'2.")
        >>> notes = repeater(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                d'16
                e'16
                f'16
                g'2.
                c'16
                d'16
                e'16
                f'16
                g'2.
            }

        ..  figure:: ../_images/Repeater-jblq28xlso.png

    :meth:`output_n`:
        This is an alias of :meth:`__call__()`. Takes an argument ``n`` for the
        number of repetitions.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> repeater = auxjad.Repeater(container)
        >>> notes = repeater.output_n(2)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Repeater-w0hd2fp2w9e.png

    :attr:`omit_time_signatures`:
        To disable time signatures altogether, initialise this class with the
        keyword argument :attr:`omit_time_signatures` set to ``True`` (default
        is ``False``), or use the :attr:`omit_time_signatures` property after
        initialisation.

        >>> container = abjad.Container(r"c'4 d'4 e'4")
        >>> repeater = auxjad.Repeater(container,
        ...                            omit_time_signatures=True,
        ...                            )
        >>> notes = repeater(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                c'4
                d'4
                e'4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/Repeater-vr4af47iwjg.png

    :attr:`force_identical_time_signatures`:
        To force time signatures in all iterations of the output, initialise
        this class with the keyword argument
        :attr:`force_identical_time_signatures` set to ``True`` (default is
        ``False``), or use the :attr:`force_identical_time_signatures` property
        after initialisation.

        >>> container = abjad.Container(r"\time 5/4 c'2. d'4 e'4")
        >>> repeater = auxjad.Repeater(container,
        ...                            force_identical_time_signatures=True,
        ...                            )
        >>> notes = repeater(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/4
                c'2.
                d'4
                e'4
                \time 5/4
                c'2.
                d'4
                e'4
                \time 5/4
                c'2.
                d'4
                e'4
            }

        ..  figure:: ../_images/Repeater-xgpndbdb0j.png

    Dynamics, slurs, and clefs:
        By default, this class automatically handles dynamics, slurs, and
        clefs, optimising their position and omitting repetitions.

        >>> container = abjad.Container(r"\clef bass f4\pp( e4) d4(")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                \clef "bass"
                f4
                \pp
                (
                e4
                )
                d4
                (
            }

        ..  figure:: ../_images/Repeater-m1ibei9s3am.png

        This is done by invoking |auxjad.mutate.reposition_clefs()|,
        |auxjad.mutate.reposition_dynamics()|, and
        |auxjad.mutate.reposition_slurs()|. Check their documentation for
        more information on how they operate.

        >>> repeater = auxjad.Repeater(container)
        >>> notes = repeater(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                \clef "bass"
                f4
                \pp
                (
                e4
                )
                d4
                (
                f4
                e4
                )
                d4
                (
                f4
                e4
                )
                d4
            }

        ..  figure:: ../_images/Repeater-scjj2uwz2p.png

        Set theproperties  :attr:`reposition_clefs`,
        :attr:`reposition_dynamics`, and :attr:`reposition_slurs` to ``False``
        to not invoke these mutations.

        >>> repeater = auxjad.Repeater(container,
        ...                            reposition_clefs=False,
        ...                            reposition_dynamics=False,
        ...                            reposition_slurs=False,
        ...                            )
        >>> notes = repeater(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                \clef "bass"
                f4
                \pp
                (
                e4
                )
                d4
                (
                \clef "bass"
                f4
                \pp
                (
                e4
                )
                d4
                (
                \clef "bass"
                f4
                \pp
                (
                e4
                )
                d4
                (
            }

        ..  figure:: ../_images/Repeater-cc39a0h84dc.png

    ..  error::

        If a container is malformed, i.e. it has an underfilled measure before
        a time signature change, this class will raise a :exc:`ValueError`
        exception.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'1")
        >>> repeater = auxjad.Repeater(container)
        ValueError: 'contents' is malformed, with an underfull measure
        preceding a time signature change
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_contents",
        "_current_window",
        "_omit_time_signatures",
        "_force_identical_time_signatures",
        "_reposition_clefs",
        "_reposition_dynamics",
        "_reposition_slurs",
        "_repeat_type",
        "_include_2x_volta_text",
    )

    ### INITIALISER ###

    def __init__(
        self,
        contents: abjad.Container,
        *,
        omit_time_signatures: bool = False,
        force_identical_time_signatures: bool = False,
        reposition_clefs: bool = True,
        reposition_dynamics: bool = True,
        reposition_slurs: bool = True,
        repeat_type: str = "unfold",
        include_2x_volta_text: bool = True,
    ) -> None:
        self.contents = contents
        self.omit_time_signatures = omit_time_signatures
        self.force_identical_time_signatures = force_identical_time_signatures
        self.reposition_clefs = reposition_clefs
        self.reposition_dynamics = reposition_dynamics
        self.reposition_slurs = reposition_slurs
        self.repeat_type = repeat_type
        self.include_2x_volta_text = include_2x_volta_text

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return abjad.lilypond(self._contents)

    def __call__(
        self,
        n: int = 1,
    ) -> abjad.Selection:
        r"""Calls the repeater process for ``n`` iterations, returning an
        |abjad.Selection|. Default ``n`` is ``1``.
        """
        if not isinstance(n, int):
            raise TypeError("first positional argument must be 'int'")
        if n < 1:
            raise ValueError("first positional argument must be a positive 'int'")
        if self._repeat_type == "unfold":
            self._repeat_unfold(n)
        elif self._repeat_type == "volta":
            self._repeat_volta(n)
        else:
            raise ValueError("'repeat_type' must be either 'unfold' or 'volta'")
        return self.current_window

    def __next__(self) -> abjad.Selection:
        r"""Calls the shuffling process for one iteration, returning an
        |abjad.Selection|.
        """
        return self.__call__()

    def __iter__(self) -> None:
        r"""Returns an iterator, allowing instances to be used as iterators."""
        return self

    ### PUBLIC METHODS ###

    def output_n(
        self,
        n: int = 1,
    ) -> abjad.Selection:
        r"""Calls the repeater process for ``n`` iterations, returning an
        |abjad.Selection|. Default ``n`` is ``1``.
        """
        return self.__call__(n)

    ### PRIVATE METHODS ###

    def _repeat_unfold(
        self,
        n: int,
    ) -> None:
        r"""Repeats a container ``n`` times."""
        dummy_container = abjad.mutate.copy(self._contents)
        for _ in range(n - 1):
            dummy_container.extend(abjad.mutate.copy(self._contents))
        if not self._force_identical_time_signatures:
            mutate.remove_repeated_time_signatures(dummy_container[:])
        if self._reposition_clefs:
            mutate.reposition_clefs(dummy_container[:])
        if self._reposition_clefs:
            mutate.reposition_dynamics(dummy_container[:])
        if self._reposition_clefs:
            mutate.reposition_slurs(dummy_container[:])
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    def _repeat_volta(
        self,
        n: int,
    ) -> None:
        r"""Adds repetition bars and ``n`` times written indication."""
        dummy_contents = abjad.mutate.copy(self._contents)
        if n == 1:
            self._current_window = dummy_contents[:]
            dummy_contents[:] = []
            return
        repeat = abjad.Repeat(repeat_count=n)
        abjad.attach(repeat, dummy_contents)
        strings = [
            r"\tweak RehearsalMark.self-alignment-X #RIGHT",
            r"\tweak RehearsalMark.break-visibility #begin-of-line-invisible",
            r'\mark \markup{\box "' + str(n) + '×"}',
        ]
        if n > 2 or (n == 2 and self._include_2x_volta_text):
            for string in strings:
                abjad.attach(
                    abjad.LilyPondLiteral(string, format_slot="after"),
                    abjad.select(dummy_contents).leaf(-1),
                )
        dummy_container = abjad.Container([abjad.mutate.copy(dummy_contents)])
        self._current_window = dummy_container[:]
        dummy_container[:] = []
        dummy_contents[:] = []

    def _get_lilypond_format(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return self.__repr__()

    @staticmethod
    def _remove_all_time_signatures(
        container: abjad.Container,
    ) -> None:
        r"""Removes all time signatures of an |abjad.Container|."""
        for leaf in abjad.select(container).leaves():
            if abjad.get.effective(leaf, abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r"""The |abjad.Container| to be shuffled."""
        return abjad.mutate.copy(self._contents)

    @contents.setter
    def contents(
        self,
        contents: abjad.Container,
    ) -> None:
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child class")
        if not abjad.select(contents).leaves().are_contiguous_logical_voice():
            raise ValueError("'contents' must be contiguous logical voice")
        if isinstance(contents, abjad.Score):
            self._contents = abjad.Container()
            for component in contents[0].components:
                self._contents.append(abjad.mutate.copy(component))
        elif isinstance(contents, abjad.Staff):
            self._contents = abjad.Container()
            for component in contents.components:
                self._contents.append(abjad.mutate.copy(component))
        elif isinstance(contents, abjad.Tuplet):
            self._contents = abjad.Container([abjad.mutate.copy(contents)])
        else:
            self._contents = abjad.mutate.copy(contents)
        dummy_container = abjad.mutate.copy(contents)
        try:
            if not get.selection_is_full(dummy_container[:]):
                mutate.close_container(self._contents)
                mutate.close_container(dummy_container)
        except ValueError as err:
            raise ValueError(
                "'contents' is malformed, with an underfull "
                "measure preceding a time signature change"
            ) from err
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    @property
    def current_window(self) -> abjad.Selection:
        r"""Read-only property, returns the previously output selection."""
        current_window = abjad.mutate.copy(self._current_window)
        if self._omit_time_signatures:
            self._remove_all_time_signatures(current_window)
        return current_window

    @property
    def omit_time_signatures(self) -> bool:
        r"""When ``True``, all time signatures will be omitted from the
        output.
        """
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(
        self,
        omit_time_signatures: bool,
    ) -> None:
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def force_identical_time_signatures(self) -> bool:
        r"""When ``True``, all time signatures will be printed in the output,
        including repeated ones .
        """
        return self._force_identical_time_signatures

    @force_identical_time_signatures.setter
    def force_identical_time_signatures(
        self,
        force_identical_time_signatures: bool,
    ) -> None:
        if not isinstance(force_identical_time_signatures, bool):
            raise TypeError("'force_identical_time_signatures' must be 'bool'")
        self._force_identical_time_signatures = force_identical_time_signatures

    @property
    def reposition_clefs(self) -> bool:
        r"""When ``True``, |auxjad.mutate.reposition_clefs()| is invoked."""
        return self._reposition_clefs

    @reposition_clefs.setter
    def reposition_clefs(
        self,
        reposition_clefs: bool,
    ) -> None:
        if not isinstance(reposition_clefs, bool):
            raise TypeError("'reposition_clefs' must be 'bool'")
        self._reposition_clefs = reposition_clefs

    @property
    def reposition_dynamics(self) -> bool:
        r"""When ``True``, |auxjad.mutate.reposition_dynamics()| is invoked."""
        return self._reposition_dynamics

    @reposition_dynamics.setter
    def reposition_dynamics(
        self,
        reposition_dynamics: bool,
    ) -> None:
        if not isinstance(reposition_dynamics, bool):
            raise TypeError("'reposition_dynamics' must be 'bool'")
        self._reposition_dynamics = reposition_dynamics

    @property
    def reposition_slurs(self) -> bool:
        r"""When ``True``, |auxjad.mutate.reposition_slurs()| is invoked."""
        return self._reposition_slurs

    @reposition_slurs.setter
    def reposition_slurs(
        self,
        reposition_slurs: bool,
    ) -> None:
        if not isinstance(reposition_slurs, bool):
            raise TypeError("'reposition_slurs' must be 'bool'")
        self._reposition_slurs = reposition_slurs

    @property
    def repeat_type(self) -> bool:
        r"""Defines the type of repeat, either ``'unfold'`` or ``'volta'``."""
        return self._repeat_type

    @repeat_type.setter
    def repeat_type(
        self,
        repeat_type: str,
    ) -> None:
        if not isinstance(repeat_type, str):
            raise TypeError("'repeat_type' must be 'str'")
        repeat_type = repeat_type.lower()
        if repeat_type not in ("unfold", "volta"):
            raise ValueError("'repeat_type' must be either 'unfold' or 'volta'")
        self._repeat_type = repeat_type

    @property
    def include_2x_volta_text(self) -> bool:
        r"""When ``True``, a written indication for the number of repeats will
        be included for ``n=2``. Otherwise, it is included only when repeated
        more than two times.
        """
        return self._include_2x_volta_text

    @include_2x_volta_text.setter
    def include_2x_volta_text(
        self,
        include_2x_volta_text: bool,
    ) -> None:
        if not isinstance(include_2x_volta_text, bool):
            raise TypeError("'include_2x_volta_text' must be 'bool'")
        self._include_2x_volta_text = include_2x_volta_text
